# -*- coding:utf-8 -*-

"""
Pages hold and organize elements.

.. moduleauthor:: Paul Wiemann <paulwiemann@gmail.com>, Johannes Brachem <jbrachem@posteo.de>
"""
import time
import logging
import string
from abc import ABCMeta, abstractproperty, abstractmethod, ABC
from builtins import object, str
from functools import reduce
from pathlib import Path
from typing import Union
from typing import Iterator

from . import alfredlog
from . import element as elm
from .element.misc import Style
from . import saving_agent
from ._core import ExpMember
from ._helper import _DictObj
from ._helper import inherit_kwargs
from .exceptions import AlfredError, ValidationError, AbortMove


@inherit_kwargs
class _PageCore(ExpMember):
    """
    Provides core functionality for pages.

    Args:
        prefix_element_names (bool): If True, the names of all input 
            elements on this page will be prefixed with the page name.
            Defaults to None. Can be defined as a class attribute.
        minimum_display_time (str): The minimal amount of time that the page
            must be displayed, before participants can move to the next
            page. Defaults to None. Can be defined as a class attribute.
        minimum_display_time_msg (str): A page-specific message to be displayed,
            if participants try to move forward before the minimum
            display time has expired. Defaults to None, which means that
            the default message defined in config.conf will be used.
            Can be defined as a class attribute.
        progress (int, float): Can be used to manually determine the 
            level of experiment progress displayed in the experiment's
            progress bar. Can be an int or float between 0 and 100.
            Defaults to 'None', which means that the experiment-wide
            setting is used (see :attr:`.ExperimentSession.progress_bar`).
            Can be defined as a class attribute.
        {kwargs}
    """

    #: If *True*, the names of all elements added to this page will
    #: receive a prefix of the page's name.
    prefix_element_names: bool = False

    #: Can be used to manually determine to 
    #: level of experiment progress displayed in the experiment's
    #: progress bar. Can be an int or float between 0 and 100.
    #: Defaults to 'None', which means that the experiment-wide
    #: setting is used (see :attr:`.ExperimentSession.progress_bar`).
    progress: int = None

    def __init__(
        self,
        prefix_element_names: bool = None,
        minimum_display_time: str = None,
        minimum_display_time_msg: str = None,
        progress: Union[int, float] = None,
        **kwargs,
    ):
        self._minimum_display_time = "0s"
        
        if minimum_display_time is not None:
            self.minimum_display_time = minimum_display_time

        if minimum_display_time_msg:
            self._minimum_display_time_msg = minimum_display_time_msg
        else:
            self._minimum_display_time_msg = None
        
        if prefix_element_names is not None:
            self.prefix_element_names = prefix_element_names
        
        if progress is not None:
            self.progress = progress

        self._data = {}
        self._is_closed = False
        self.show_times = []
        self.hide_times = []

        self.elements = {}

        self._element_name_counter = 1

        super().__init__(**kwargs)

    @property
    def elements(self):
        """
        Dictionary of elements belonging to this page.
        """
        return self._elements
    
    @elements.setter
    def elements(self, value):
        self._elements = value

    @property
    def minimum_display_time(self) -> str:
        """
        str: Minimal amount of time that a page must be displayed
        before participants can move forward.

        Must be specified as a string with a unit of 's' (for seconds),
        or 'm' (for minutes).
        """
        return self._minimum_display_time
    
    @property
    def _mdt(self) -> float:
        mdt = self.minimum_display_time
        unit = mdt[-1]
        if unit == "s":
            return float(mdt[:-1])
        elif unit == "m":
            return float(mdt[:-1]) * 60
        else:
            raise ValueError(
                "Please specify minimum display time with a unit ('s' - seconds, 'm' - minutes)."
            )

    @minimum_display_time.setter
    def minimum_display_time(self, value: str):
        self._minimum_display_time = value

    def added_to_experiment(self, experiment):
        # docstring inherited
        super().added_to_experiment(experiment)
        self.log.add_queue_logger(self, __name__)

        debug = self.experiment.config.getboolean("general", "debug")
        if debug and not self._mdt == 0:
            if self.experiment.config.getboolean("debug", "disable_minimum_display_time"):
                self.log.debug("Minimum display time disabled (debug mode).")
                self.minimum_display_time = "0s"

    @property
    def minimum_display_time_msg(self) -> str:
        """
        Message that is displayed if participants try to leave a page
        before the minimum display time is up.
        """
        msg = self._minimum_display_time_msg
        if msg is not None:
            return msg
        else:
            return self.experiment.config.get("hints", "minimum_display_time")

    @minimum_display_time_msg.setter
    def minimum_display_time_msg(self, value: str):
        self._minimum_display_time_msg = value

    @property
    def is_closed(self) -> bool:
        """
        Returns *True*, if the page is closed.
        """
        return self._is_closed

    @property
    def should_be_shown(self) -> bool:
        """
        bool: Boolean, indicating whether a page should be shown.

        Evaluates the page's own settings, as well as the status of all
        of its parent sections. The page is only shown, if all
        conditions evaluate to *True*.
        """
        thispage = super().should_be_shown
        sections = [sec.should_be_shown for sec in self.uptree()]
        return thispage and all(sections)

    @should_be_shown.setter
    def should_be_shown(self, value: bool):
        self._should_be_shown = bool(value)

    @property
    def must_be_shown(self) -> bool:
        """
        bool: False, if the experiment tolerates skipping this
        page entirely. Defaults to False.

        Notes:
            If there are input elements with *force_input=True* on a
            skippable page, particpants will be notified and validation
            will fail, even if the *must_be_shown* attribute of the page
            itself is set to False.

            If a page that must be shown has not been shown, the
            experiment will display the hint *page_must_be_shown* as defined
            in config.conf, section "hints".

            ..warning::
                It is easy to end up in a situation where
                a mandatory page has not been shown, but the participant has
                no way of moving back to it. So, please be careful when
                using this option.

        """
        return False

    @property
    def has_been_shown(self) -> bool:
        """
        Returns *True*, if the page has been displayed in the ongoing
        experiment session.
        """
        return self._has_been_shown

    def _on_showing_widget(self, show_time: float = None):
        """
        Method for internal processes on showing Widget

        Args:
            show_time: Time of showing in seconds since epoch.
        """
        if show_time is None:
            show_time = time.time()
        self.show_times.append(show_time)

        if not self._has_been_shown:
            self.on_first_show()

            if self.exp.config.getboolean("general", "debug") and self is not self.exp.final_page:
                name = self.name + "__debug_jumplist__"
                jumplist = elm.action.JumpList(
                    scope="exp",
                    check_jumpto=False,
                    check_jumpfrom=False,
                    name=name,
                    debugmode=True,
                )
                jumplist.should_be_shown = False
                self += jumplist

        self.on_each_show()

        self._has_been_shown = True
        if self.exp.aborted:
            raise AbortMove

    def on_first_show(self):
        """
        Executed *once*, when the page is shown for the first time,
        *before* executing :meth:`~.Page.on_each_show`.

        This is your go-to-hook, if you want to have access to data
        from other pages within the experiment, and your code is meant
        to be executed only once (i.e. the first time a page is shown).

        See Also:
            See "How to use hooks" for a how to on using hooks and an overview
            of available hooks.

        """
        pass

    def on_each_show(self):
        """
        Executed *every time* the page is shown, *after* executing
        :meth:`~.Page.on_first_show`.

        See Also:
            See "How to use hooks" for a how to on using hooks and an overview
            of available hooks.

        """
        pass

    def _on_hiding_widget(self, hide_time: float = None):
        """
        Method for internal processes on hiding Widget

        Args:
            hide_time: Time of hiding in seconds since epoch.
        """
        if hide_time is None:
            hide_time = time.time()

        self.hide_times.append(hide_time)

        if not self._has_been_hidden:
            self.on_first_hide()

        self.on_each_hide()

        self._has_been_hidden = True

        self.save_data()

        if self.exp.aborted:
            raise AbortMove

    def on_first_hide(self):
        """
        Executed *once*, when the page is hidden for the first time,
        *before* executing :meth:`~.Page.on_each_hide`.

        Hook for code that is meant to be executed only once, when
        the page is hidden for the first time, **before** saving the
        page's data.

        Notes:
            Note the difference to :meth:`on_close`, which is
            executed upon final submission of the page's data. When using
            :meth:`on_first_hide`, subject input can change (e.g., when a
            subject revists a page and changes his/her input).

        See Also:
            See "How to use hooks" for a how to on using hooks and an overview
            of available hooks.

        """
        pass

    def on_each_hide(self):
        """
        Executed *every time* the page is hidden, *before* closing it
        and *before* saving data, but *after* executing
        :meth:`~.Page.on_first_hide`.

        See Also:
            See "How to use hooks" for a how to on using hooks and an overview
            of available hooks.

        """
        pass

    def on_close(self):
        """
        Executed *once*, when the page is closed, *before* data saving.

        This is your go-to-hook, if you want to have the page execute
        this code only once, when submitting the data from a page. After
        a page is closed, there can be no more changes to subject input.
        This is the most important difference of :meth:`on_close` from
        :meth:`on_first_hide` and :meth`on_each_hide`.

        See Also:
            See "How to use hooks" for a how to on using hooks and an overview
            of available hooks.

        """
        pass

    def on_exp_access(self):
        """
        Executed *once*, when the :class:`.ExperimentSession` becomes
        available to the page.

        This is your go-to-hook, if you want to have access to the
        experiment, but don't need access to data from other pages.

        See Also:
            See "How to use hooks" for a how to on using hooks and an overview
            of available hooks.

        """
        pass

    def close(self):
        """
        Closes the page. Usually, sections will take care of closing
        pages. Only call this method, if you cannot solve your problem
        by using a certain section.
        """
        self.on_close()
        self._is_closed = True

    def validate_page(self) -> bool:
        """
        Returns *True*, if the validation checks pass, *False* otherwise.
        """
        return True

    def validate_elements(self) -> bool:
        """
        Returns *True*, if validation of all input elements passes,
        *False* otherwise.
        """
        return True

    def save_data(self, level: int = 1, sync: bool = False):
        """
        Saves current experiment data.

        Collects the current experimental data and calls the
        experiment's main SavingAgentController to save the data with
        all SavingAgents.

        Args:
            level: Level of the saving task. High level means high
                importance. If the level is below a SavingAgent's
                activation level, that agent will not be used for
                processing this task. Defaults to 1.
            sync: If True, the saving task will be prioritised and the
                experiment will pause until the task was fully completed.
                Should be used carefully. Defaults to False.

        Notes:
            Note that this function will be called automatically on every
            move in the experiment. Only call it yourself if it is 
            really necessary.
        """
        if not self.exp.data_saver.main.agents and not self.exp.config.getboolean(
            "general", "debug"
        ):
            self.log.warning("No saving agents available.")
        data = self.experiment.data_manager.session_data
        self.exp.data_saver.main.save_with_all_agents(data=data, level=level, sync=sync)

    def __repr__(self):
        return f"Page(class='{type(self).__name__}', name='{self.name}')"

    def prepare_web_widget(self):
        """
        Hook for computations for preparing a page for display.

        Gets executed *every time* before the page is displayed.
        """
        pass

    @property
    def _css_code(self):
        return []

    @property
    def _css_urls(self):
        return []

    @property
    def _js_code(self):
        return []

    @property
    def _js_urls(self):
        return []

    def _set_data(self, dictionary: dict):
        """
        Informs the page's input elements about user input.
        """
        pass


@inherit_kwargs
class _CoreCompositePage(_PageCore):
    """
    Second base class for pages.

    Args:
        {kwargs}
    """

    def __contains__(self, element):
        try:
            return element.name in self.elements
        except AttributeError:
            return element in self.elements

    # necessary to make __getattr__ work with copying a page object
    def __getstate__(self):
        return self.__dict__

    # necessary to make __getattr__ work with copying a page object
    def __setstate__(self, state):
        self.__dict__.update(state)

    def __getitem__(self, name):
        return self.elements[name]

    def __getattr__(self, name):
        try:
            return self.elements[name]
        except KeyError:
            raise AttributeError(f"{self} has no attribute '{name}'.")

    @property
    def input_elements(self) -> dict:
        """
        Dict of all input elements on this page.

        Does not evaluate whether an input element should be shown or
        not, because that might change over the course of an experiment.
        """

        input_elements = {}
        for name, el in self.elements.items():
            if isinstance(el, (elm.core.InputElement)):
                input_elements[name] = el
        return input_elements

    @property
    def all_input_elements(self) -> dict:
        """
        Alias for :attr:`.input_elements`.
        """
        return self.input_elements

    @property
    def all_elements(self) -> dict:
        """
        Alias for :attr:`.elements`.
        """
        return self.elements

    @property
    def updated_elements(self) -> dict:
        """
        Returns a dict of all elements on the page that already have 
        access to the experiment session.
        """
        return {name: elm for name, elm in self.elements.items() if elm.exp is not None}

    @property
    def filled_input_elements(self) -> dict:
        """
        Dict of all input elements on this page with non-empty data 
        attribute. This includes elements that hav received participant
        input and elements with default values.
        """

        return {name: el for name, el in self.input_elements.items() if el.input}

    def append(self, *elements):
        """
        Appends a variable number of elements to the page.

        In practice, it is recommended to use the augmented assignment
        operator ``+=`` instead in order to add elements.
        """
        for elmnt in elements:
            if not isinstance(elmnt, (elm.core.Element)):
                raise TypeError(f"Can only append elements to pages, not '{type(elmnt).__name__}'")

            elmnt.added_to_page(self)

            if elmnt.name in dir(self):
                raise ValueError(f"Element name '{elmnt.name}' is also an attribute of {self}.")

            if elmnt.name in self.elements:
                raise AlfredError(f"{self} already has an element of name '{elmnt.name}'.")

            if self.exp is not None and elmnt.exp is None:
                elmnt.added_to_experiment(self.exp)

            self.elements[elmnt.name] = elmnt

    def _generate_element_name(self, element):
        i = self._element_name_counter
        c = element.__class__.__name__
        self._element_name_counter += 1

        return f"{self.name}_{c}_{i}"

    def __iadd__(self, other):
        self.append(other)
        return self


    def added_to_experiment(self, experiment):
        # docstring inherited
        super().added_to_experiment(experiment)
        self.on_exp_access()
        self._update_elements()

    def added_to_section(self, section):
        # docstring inherited
        super().added_to_section(section)
        self._update_elements()

    def _update_members_recursively(self):
        self._update_elements()

    def _update_elements(self):
        if self.exp and self.section and self.tree.startswith("_root"):
            for element in self.elements.values():
                if not element.exp:
                    element.added_to_experiment(self.experiment)

    def close(self):
        # docstring inherited
        super().close()

        for elmnt in self.elements.values():
            if isinstance(elmnt, elm.core.InputElement):
                elmnt.disabled = True

        debug_jumplist = self.elements.get(self.name + "__debug_jumplist__")
        if debug_jumplist:
            for elmnt in debug_jumplist.elements:
                elmnt.disabled = False

    @property
    def data(self) -> dict:
        """
        Returns a dict of data for all input elements on the page. 

        If the page has not been shown yet, an empty dict is returned.
        """

        if not self.has_been_shown:
            return {}
        else:
            data = {}
            for element in self.input_elements.values():
                data.update(element.data)
            return data

    @property
    def unlinked_data(self) -> dict:
        """
        Returns an empty dict for 'normal' pages and the input data
        for unlinked pages.
        """
        return {}

    def _set_data(self, dictionary: dict):
        for elmnt in self.input_elements.values():
            elmnt.set_data(dictionary)

    def custom_move(self):
        """
        Hook for defining a page's own movement behavior, executed
        *every time* a movement *from* the page takes place,
        *before* :meth:`~.Page.on_first_hide` and :meth:`~.Page.on_each_hide`.

        User input to the elements on the current page is available in
        this method through the page's :attr:`.Page.data` attribute.

        Use the :class:`.ExperimentSession` s movement methods to define
        your own behavior. The available methods are

        .. autosummary::
           :nosignatures:

           ~alfred3.experiment.ExperimentSession.forward
           ~alfred3.experiment.ExperimentSession.backward
           ~alfred3.experiment.ExperimentSession.jump

        Notes:
            You can fall back to alfred3's movement system by returning
            *True* from your custom move function.

        Examples:

            Create a page that always jumps to a specific page upon
            submission::

                exp = al.Experiment()

                @exp.member
                class CustomMove(al.Page):
                    name = "custom_move"

                    def custom_move(self):
                        self.exp.jump(to="third")

                exp += al.Page(name="second")
                exp += al.Page(name="third")


            Create a page that jumps to a specific page, if it received
            a user input of 'yes', and use alfred3's usual movement
            system otherwise::

                exp = al.Experiment()

                @exp.member
                class CustomMove(al.Page):

                    def on_exp_access(self):
                        self += elm.TextEntry(name="text")

                    def custom_move(self):
                        if self.data.get("text) == "yes":
                            self.exp.jump(to="third")
                        else:
                            return True

                exp += al.Page(name="second")
                exp += al.Page(name="third")

        """
        return True

    def durations(self) -> Iterator[float]:
        """
        Iterates over the visit durations for this page. 
        
        Yields:
            float: Duration of a visit in seconds.
        
        """

        if len(self.show_times) > len(self.hide_times):
            now = time.time()
        elif len(self.show_times) < len(self.hide_times):
            self.log.error(f"{self} has fewer entries in show_times than in hide_times.")

        for show, hide in zip(self.show_times, self.hide_times + [now]):
            yield hide - show

    def last_duration(self) -> float:
        """
        Returns the duration of the last visit to this page in the 
        current session in seconds.
        """
        *_, last_duration = self.durations()
        return last_duration

    def first_duration(self) -> float:
        """
        Returns the duration of the last visit to this page in the 
        current session in seconds.
        """
        first_duration, *_ = self.durations()
        return first_duration

    def _validate_page(self):

        if not self.has_been_shown:
            if self.must_be_shown:
                msg = self.exp.config.get("hints", "page_must_be_shown")
                self.exp.post_message(msg, level="danger")
                return False
            else:
                return True

        # check minimum display time
        mintime = self._mdt
        if time.time() - self.show_times[0] < mintime:
            msg = self.minimum_display_time_msg.format(mdt=str(mintime))
            self.exp.message_manager.post_message(msg)
            return False

        return True

    def _validate_elements(self):
        return all([el.validate_data() for el in self.input_elements.values()])


@inherit_kwargs
class Page(_CoreCompositePage):
    """
    The basic page.

    Args:
        fixed_width (str): Custom value for defining a fixed width of
            the page. Only takes effect, if the experiment is set to
            generally operate with a fixed page width in *config.conf*
            (option *responsive* in section *layout* must be "false").
            Must be a string including a unit, e.g. "900px". Can be
            defined as a class attribute.

        responsive_width (str): Custom values for definig the width of
            the page in percent of the screen width. Only takes
            effect, if the option *responsive* in section *layout* is
            "true" (which is the default). Must be a single string with
            1 to 5 relative widths separated by commas, e.g. "60%, 50%".
            The first value refers to extra small ('xs') screens, the
            following values to the next bigger ones. If the string
            contains less than five values, the last value will be used
            for all screens from that size on upward.

            The sizes are taken from Bootstrap and correspond to the
            five width attributes of a :class:`.RowLayout`.

            Can be defined as a class attribute.

        header_color (str): A color to be used for the header
            of this page. Can be any color value understood by CSS,
            including hex and RGB. Can be defined as a class
            attribute.

        background_color (str): A color to be used for the background of
            this page. Can be any color value understood by CSS,
            including hex and RGB. Can be defined as a class
            attribute.

        {kwargs}

    Notes:

        .. note:: In class style, you can use the initialization 
           arguments of a page by defining them as class attributes.
           See "How to use pages" for details.
    
    Examples:

        Basic example for adding a page with a single element in
        class style::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class DemoPage(al.Page):
                title = "This is a demo"

                def on_exp_access(self):
                    self += al.Text("This is demo text.")
            

        Example for customly defining a set of relative widths. The 
        following page will take up 80 % of available space on extra 
        small screens, 70 % on small screens, and 60 % on medium, large 
        and extra large screens::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                responsive_width = "80%, 70%, 60%"

    """

    def __init__(
        self,
        title: str = None,
        name: str = None,
        fixed_width: str = None,
        responsive_width: str = None,
        header_color: str = None,
        background_color: str = None,
        *args,
        **kwargs,
    ):
        super().__init__(title=title, name=name, *args, **kwargs)

        self._fixed_width = None
        if fixed_width:
            self.fixed_width = fixed_width

        self._responsive_width = None
        if responsive_width:
            self.responsive_width = responsive_width

        self._header_color = None
        if header_color:
            self.header_color = header_color

        self._background_color = None
        if background_color:
            self.background_color = background_color

    @property
    def fixed_width(self) -> str:
        """
        Custom value for defining a fixed width of
        the page. Only takes effect, if the experiment is set to
        generally operate with a fixed page width in *config.conf*
        (option *responsive* in section *layout* must be "false").
        Must be a string including a unit, e.g. "900px". Can be
        defined as a class attribute.
        """
        return self._fixed_width

    @fixed_width.setter
    def fixed_width(self, value):
        self._fixed_width = value

    @property
    def responsive_width(self):
        """
        Custom values for definig the width of
        the page in percent of the screen width. Only takes
        effect, if the option *responsive* in section *layout* is
        "true" (which is the default). Must be a single string with
        1 to 5 relative widths separated by commas, e.g. "60%, 50%".
        The first value refers to extra small ('xs') screens, the
        following values to the next bigger ones. If the string
        contains less than five values, the last value will be used
        for all screens from that size on upward.

        The sizes are taken from Bootstrap and correspond to the
        five width attributes of a :class:`.RowLayout`.

        Can be defined as a class attribute.
        """
        return self._responsive_width

    @responsive_width.setter
    def responsive_width(self, value):
        self._responsive_width = value

    @property
    def header_color(self):
        """
        A color to be used for the header
        of this page. Can be any color value understood by CSS,
        including hex and RGB. Can be defined as a class
        attribute.
        """
        return self._header_color

    @header_color.setter
    def header_color(self, value):
        self._header_color = value

    @property
    def background_color(self):
        """
        A color to be used for the background of
        this page. Can be any color value understood by CSS,
        including hex and RGB. Can be defined as a class
        attribute.
        """
        return self._background_color

    @background_color.setter
    def background_color(self, value):
        self._background_color = value

    def _parse_responsive_width(self, width):
        return [x.strip().replace("%", "") for x in width.split(",")]

    def _responsive_media_query(self, width):
        if len(width) > 4:
            raise ValueError("The option 'responsive_width' can only define up to four widths.")

        if len(width) < 4:
            for _ in range(4 - len(width)):
                width.append(width[-1])

        screen_size = [576, 768, 992, 1200]
        t = string.Template(
            "@media (min-width: ${screen}px) {.responsive-width { width: ${w}%; max-width: none;}}"
        )
        out = []
        for i, w in enumerate(width):
            out.append(t.substitute(screen=screen_size[i], w=w))
        return " ".join(out)

    def prepare_web_widget(self):
        # docstring inherited
        for elmnt in self.elements.values():
            elmnt._prepare_web_widget()

    def added_to_experiment(self, experiment):
        # docstring inherited
        super().added_to_experiment(experiment)
        self._set_width()
        self._set_color()

    def _set_width(self):
        if self.experiment.config.getboolean("layout", "responsive"):

            if self.responsive_width:
                w = self._parse_responsive_width(self.responsive_width)
                self += Style(code=self._responsive_media_query(w))

            elif self.experiment.config.get("layout", "responsive_width"):
                config_width = self.experiment.config.get("layout", "responsive_width")
                w = self._parse_responsive_width(config_width)
                self += Style(code=self._responsive_media_query(w))

        elif not self.fixed_width:
            w = self.experiment.config.get("layout", "fixed_width")
            self += Style(code=f".fixed-width {{ width: {w}; }}")
            self += Style(code=f".min-width {{ min-width: {w}; }}")

        else:
            self += Style(code=f".fixed-width {{ width: {self.fixed_width}; }}")
            self += Style(code=f".min-width {{ min-width: {self.fixed_width}; }}")

    def _set_color(self):
        if self.header_color:
            self += Style(code=f".logo-bg {{background-color: {self.header_color};}}")

        elif self.experiment.config.get("layout", "header_color", fallback=False):
            c = self.experiment.config.get("layout", "header_color", fallback=False)
            self += Style(code=f".logo-bg {{background-color: {c};}}")

        if self.background_color:
            self += Style(code=f"body {{background-color: {self.background_color};}}")

        elif self.experiment.config.get("layout", "background_color", fallback=False):
            c = self.experiment.config.get("layout", "background_color", fallback=False)
            self += Style(code=f"body {{background-color: {c};}}")

    @property
    def _css_code(self):
        return reduce(lambda l, element: l + element.css_code, self.elements.values(), [])

    @property
    def _css_urls(self):
        return reduce(lambda l, element: l + element.css_urls, self.elements.values(), [])

    @property
    def _js_code(self):
        return reduce(lambda l, element: l + element.js_code, self.elements.values(), [])

    @property
    def _js_urls(self):
        return reduce(lambda l, element: l + element.js_urls, self.elements.values(), [])

@inherit_kwargs
class WidePage(Page):
    """
    A page with a wider default width.

    Args:
        {kwargs}

    Notes:
        The width of this page is only customized for responsive layouts.
        If you disabled responsive layout, it will have default width.
    
    Examples:
        A minimal experiment with a single wide page, holding a text
        element::
            
            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class DemoPage(al.WidePage):
                title = "This is a demo"

                def on_exp_access(self):
                    self += al.Text("This is demo text.")
    
    """

    responsive_width = "85%, 75%, 75%, 70%"

@inherit_kwargs
class NoNavigationPage(Page):
    """
    A page without navigation buttons.
    
    Args:
        {kwargs}
    
    Examples:
        A minimal experiment with a single NoNavigationPage, from which
        participants can move forward via the SubmittingButtons::
            
            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class DemoPage(al.NoNavigationPage):
                title = "This is a demo"

                def on_exp_access(self):
                    
                    self += al.SubmittingButtons(
                        "Yes", "No", 
                        toplab="Do you agree to this statement?",
                        name="submit1"
                        )
    """

    def added_to_experiment(self, experiment):
        super().added_to_experiment(experiment)
        self += Style("#page-navigation {display: none;}")


@inherit_kwargs
class TimeoutPage(Page):
    """
    A page with an additional hook that gets triggered after a timeout.
    
    The timeout starts when the page is shown for the first time. If a
    page is shown for a second time, the timeout will not run again.

    Args:
        timeout (str): Length of the timeout. Specify it as a string 
            with a unit of "s" for seconds, and "m" for minutes, 
            for example "10s". Can be specified as a class attribute.
        {kwargs}
    
    See Also:
        Timeout hook: :meth:`~.Page.on_timeout`
    
    Examples:
        A minimal experiment with a single TimeoutPage, holding a text
        element. The navigating buttons are hidden through the 
        :class:`.HideNavigation` element.
        
        The *on_timeout* hook shows the implementation of the
        :class:`.AutoForwardPage`. The page will move forward after the
        timeout of 5 seconds expires::
            
            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class DemoPage(al.TimeoutPage):
                title = "This is a demo"
                timeout = "5s"

                def on_exp_access(self):
                    self += al.Text("This is demo text.")
                    self += al.HideNavigation()
                
                def on_timeout(self):
                    self.exp.forward()
    """

    timeout = None

    def __init__(self, timeout: str = None, **kwargs):
        super().__init__(**kwargs)

        self._end_link = "unset"
        self._run_timeout = True
        if timeout is not None:
            self.timeout = timeout

        if self.timeout is None:
            raise AlfredError("A TimeoutPage must have a 'timeout' attribute.")

        unit = self.timeout[-1]
        if unit == "s":
            self.timeout = int(self.timeout[:-1])
        elif unit == "m":
            self.timeout = int(self.timeout[:-1]) * 60
        else:
            raise ValueError(
                "You must specify the unit of your timeout ('s' - seconds, or 'm' - minutes)"
            )

    def added_to_experiment(self, experiment):
        # docstring inherited
        super().added_to_experiment(experiment)
        self._end_link = self._experiment.user_interface_controller.add_callable(self._callback)

        if self._experiment.config.getboolean("general", "debug"):
            if self._experiment.config.getboolean("debug", "reduce_countdown"):
                self.timeout = self._experiment.config.getint("debug", "reduced_countdown_time")

    @property
    def _js_code(self):
        code = (
            5,
            """
            $(document).ready(function(){
                var start_time = new Date();
                var timeout = %s;
                var action_url = '%s';

                var update_counter = function() {
                    var now = new Date();
                    var time_left = timeout - Math.floor((now - start_time) / 1000);
                    if (time_left < 0) {
                        time_left = 0;
                    }
                    $(".timeout-label").html(time_left);
                    if (time_left > 0) {
                        setTimeout(update_counter, 200);
                    }
                };
                update_counter();

                var timeout_function = function() {
                    $("#form").attr("action", action_url);
                    $("#form").submit();
                };
                setTimeout(timeout_function, timeout*1000);
            });
        """
            % (self.timeout, self._end_link),
        )
        js_code = super()._js_code
        if self._run_timeout:
            js_code.append(code)
        else:
            js_code.append((5, """$(document).ready(function(){$(".timeout-label").html(0);});"""))
        return js_code

    def _callback(self, **kwargs):
        self._run_timeout = False
        self._experiment.movement_manager.current_page._set_data(kwargs)
        self.on_timeout()

    def on_timeout(self):
        """
        Executed *once*, after the timeout of the page runs out.

        See Also:
            See "How to use hooks" for a how to on using hooks and an overview
            of available hooks.

            This hook is defined by :class:`.AutoForwardPage` and
            :class:`.AutoClosePage`.
        """
        pass


@inherit_kwargs
class AutoForwardPage(TimeoutPage):
    """
    A page that automatically moves forward after the timeout expires.

    Args:
        {kwargs}

    Notes:
        This page will work with customly defined :meth:`.custom_move`
        methods.

    Examples:
        ::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class DemoPage(al.AutoForwardPage):
                timeout = "5s"

                def on_exp_access(self):
                    self += al.Text("This page will move after 5 seconds.")

    """

    def on_timeout(self):
        self.experiment.movement_manager.move(direction="forward")


@inherit_kwargs
class AutoClosePage(TimeoutPage):
    """
    A page that automatically closes itself after the timeout expires.

    Args:
        {kwargs}

    Examples:
        ::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class DemoPage(al.AutoClosePage):
                timeout = "5s"

                def on_exp_access(self):
                    self += al.Text("This page will close after 5 seconds.")

    """

    def on_timeout(self):
        self.close()


@inherit_kwargs
class NoDataPage(Page):
    """
    A page that does not save any data.

    Args:
        {kwargs}
    
    Notes:
        This page *will* still trigger a saving event upon moving,
        so that, for example, updates to the :attr:`.ExperimentSession.adata` 
        dictionary will be saved. It just does not save any data itself,
        i.e. input elements on this page will not appear in the data.
    
    See Also:
        See :class:`.NoSavingPage` for a page that does not trigger a
        saving event.
    """

    data = {}

@inherit_kwargs
class NoSavingPage(Page):
    """
    A page that does not trigger a saving event on moving.

    Args:
        {kwargs}
    
    Notes:
        This page will not trigger any saving action on its own. But if
        experiment data is saved at another point of the experiment, data
        collected on a NoSavingPage *will* be saved. You can switch off 
        this behavior by overriding the *data* attribute, effectively
        creating a NoDataNoSavingPage (see Example 2).

        Note also that upon finishing, the experiment will always trigger
        a saving event, even if you only have NoSavingPages in your
        experiment. To prevent this, you can switch the option 
        'save_data' in section 'data' of config.conf to 'false'.

    See Also:
        See :class:`.NoDataPage` for a page that does not collect any
        data.
    
    Examples:

        Example 1, simple usage::

            import alfred3 as al
            exp = al.Experiment()
            exp += al.NoSavingPage(name="demo")

        Example 2, a NoSavingPage that also does not collect any data::

            import alfred3 as al
            exp = al.Experiment()

            class NoDataNoSavingPage(al.NoSavingPage):

                data = {{}}
            
            exp += NoDataNoSavingPage(name="demo")
        
        Example 3, a more elaborate example. Here, we create an 
        experiment with an "admin" mode, triggered by an url parameter,
        in which no data will be saved.
        Note that, if we use this method, we do not even need to use
        a NoSavingPage::

            import alfred3 as al
            exp = al.Experiment()

            @exp.setup
            def setup(exp):
                admin = exp.urlargs.get("admin", False)
                if admin == "true":
                    exp.config.read_dict({{"data": {{"save_data": "false"}}}})
                    exp.log.info("Admin mode triggered. No data will be saved")

            exp += al.Page(name="demo")
        
        If this experiment is started with the suffix ``?admin=true`` to
        the starting url, no data will be saved. In case of a locally 
        running experiment, the url would be 
        http://localhost:5000/start?admin=true

        .. note:: Note that the experiment will most likely crash, if
            the /start route gets called twice for any reason in a local
            experiment.

    """

    def save_data(self, *args, **kwargs):
        """
        This page type does not save data on its own, so this method has
        no effect.
        """
        pass


@inherit_kwargs
class UnlinkedDataPage(NoDataPage):
    """
    A page that saves data separately from the experiment data.

    Args:
        encrypt (str): Takes one the following values: 'agent' (default) will
            encrypt data based on each saving agent's configuration.
            'always' will encrypt all data entered on this page,
            regardless of saving agent configuration. 'never' will turn
            off encryption for this page, regardless of saving agent
            configuration. Can be specified as a class attribute.
        {kwargs}

    Unlinked data is data that does not contain any identifiers that
    would allow someone to establish a connection between an
    experiment data set and the unlinked data set.
    A common use case is the sepration of identifying personal
    information that might be needed for non-experimental purposes such
    as compensation admninistration, from experiment data.

    In practice that means that the UnlinkedDataPage does not save any
    of the following:

    - Time of saving
    - Session ID
    - Experiment condition
    - Additional data
    - Start time
    - Experiment version
    - Alfred version

    It will, however, save the following information:

    - Experiment Title
    - Experiment ID
    - Experiment Author
    - Page tag
    - Page ID

    Thus, the saved data *can* be linked to an *experiment* and to a
    page. That is intended and indeed necessary so that data
    can be retrieved and processed. The key point is that there is no
    identifier for linking data to data from a specific experimental
    *session* (i.e. the name of a subject, saved with an
    UnlinkedDataPage cannot be linked to his/her answers given on other
    Pages).

    .. warning::
        All data from UnlinkedDataPages is saved in a single unlinked
        data document, so data from two different unlinked pages *are*
        linked to each other (though not to the rest of the experiment
        data).

    """

    encrypt: str = "agent"

    def __init__(self, encrypt: str = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        if encrypt is not None:
            self.encrypt = encrypt

        if self.encrypt not in ["agent", "always", "never"]:
            raise ValueError(
                "The argument 'encrypt' must take one of the following values: 'agent', 'always', 'never'."
            )

    @property
    def unlinked_data(self) -> dict:
        # docstring inherited
        if not self.has_been_shown:
            return {}
        else:
            data = {}
            for element in self.input_elements.values():
                data.update(element.data)
            return data

    def save_data(self, level: int = 1, sync: bool = False):
        """
        Saves current unlinked data.

        Collects the unlinked data from all UnlinkedDataPages in the
        experiment and engages the experiment's unlinked
        SavingAgentController to save the data with all SavingAgents.

        Args:
            level: Level of the saving task. High level means high
                importance. If the level is below a SavingAgent's
                activation level, that agent will not be used for
                processing this task. Defaults to 1.
            sync: If True, the saving task will be prioritised and the
                experiment will pause until the task was fully completed.
                Should be used carefully. Defaults to False.
        """
        if (
            not self._experiment.data_saver.unlinked.agents
            and not self._experiment.config.getboolean("general", "debug")
        ):
            self.log.warning("No saving agent for unlinked data available.")

        for agent in self._experiment.data_saver.unlinked.agents.values():

            if self.encrypt == "agent":
                data = self.experiment.data_manager.unlinked_data_with(agent)
            elif self.encrypt == "always":
                data = self.experiment.data_manager.unlinked_data
                data = self.experiment.data_manager.encrypt_values(data)
            elif self.encrypt == "never":
                data = self.experiment.data_manager.unlinked_data

            self.exp.data_saver.unlinked.save_with_agent(
                data=data, name=agent.name, level=level, sync=sync
            )


class _CustomSavingPage(Page, ABC):
    """
    TODO: This class still needs some work.

    Allows you to add custom SavingAgents directly to the page.

    Args:
        save_to_main: If True, data will be *also* saved using the
            experiment's main SavingAgentController and all of its
            SavingAgents. Defaults to False. Can be specified as a
            class attribute.

    Notes:
        Since this is an abstract class, it can not be instantiated directly.
        You have to derive a child class and define the property
        :meth:`custom_save_data`, which must return a dictionary. Through
        this property, you control exactly which data will be saved by this
        page.

        .. warning::
            Each SavingAgent maintains one file or one document.
            On saving, the document will be fully replaced with the current
            data. That means, you should not let two CustomSavingPages
            share a SavingAgent, as they will override each other's data.
            That is, unless that is your intended behavior, e.g. when the
            pages share data.

    Examples:
        Example 1: Saving ordinary page data (like other pages)::

            class MyPage(CustomSavingPage):

                @property
                def custom_save_data(self):
                    return self.data


        Example 2: Saving a static dictionary::

            class MyPage(CustomSavingPage):

                @property
                def custom_save_data(self):
                    return {"key": "value"}
    """

    save_to_main: bool = False

    def __init__(self, save_to_main: bool = None, **kwargs):
        super().__init__(**kwargs)
        
        if save_to_main is not None:
            self.save_to_main = save_to_main

    def added_to_experiment(self, experiment):
        if not self._experiment:
            super().added_to_experiment(experiment=experiment)
            self.saving_agent_controller = saving_agent.SavingAgentController(self._experiment)
        self._check_for_duplicate_agents()

    def _check_for_duplicate_agents(self):
        comparison = []
        comparison += list(self._experiment.data_saver.main.agents.values())
        comparison += list(self._experiment.data_saver.unlinked.agents.values())

        for pg in self._experiment.root_section.all_pages.values():
            if pg == self:
                continue
            try:
                comparison += list(pg.saving_agent_controller.agents.values())
            except AttributeError:
                pass

        for agent in self.saving_agent_controller.agents.values():
            self._check_one_duplicate(agent, comparison)

    @staticmethod
    def _check_one_duplicate(agent, agents_list):
        for ag in agents_list:
            if ag == agent:
                raise ValueError("A SavingAgent added to a CustomSavingPage must be unique")

    def append_saving_agents(self, *args):
        """
        Appends saving agents to this page.

        These saving agents will be used to save the page's data.
        """
        for agent in args:
            self.saving_agent_controller.append(agent)
        self._check_for_duplicate_agents()

    def append_failure_saving_agents(self, *args):
        for agent in args:
            self.saving_agent_controller.append_failure_agent(agent)

    @abstractproperty
    def custom_save_data(self):
        pass

    def save_data(self, level=1, sync=False):

        if not isinstance(self.custom_save_data, dict):
            raise ValueError("The porperty 'custom_page_data' must return a dictionary.")

        if self.save_to_main:
            self._experiment.data_saver.main.save_with_all_agents(level=level, sync=sync)

        self.saving_agent_controller.save_with_all_agents(
            data=self.custom_save_data, level=level, sync=sync
        )


class _DefaultFinalPage(Page):
    """
    The default final page.
    """

    title = "Experiment beendet"

    def on_exp_access(self):
        txt = "Das Experiment ist nun beendet.<br>Vielen Dank für die Teilnahme."
        self += elm.display.Text(":mortar_board:", font_size=70, align="center")
        self += elm.display.VerticalSpace("20px")
        self += elm.display.Text(text=txt, align="center")
        self += elm.misc.WebExitEnabler()



# -*- coding: utf-8 -*-
from ._version import __version__

from .experiment import Experiment

from .condition import ListRandomizer
from .condition import AllConditionsFull
from .condition import random_condition

from .section import Section
from .section import RevisitSection
from .section import ForwardOnlySection

from .page import Page
from .page import WidePage
from .page import UnlinkedDataPage
from .page import AutoForwardPage
from .page import AutoClosePage
from .page import NoNavigationPage
from .page import NoDataPage
from .page import NoSavingPage

from .element.core import Row
from .element.core import Stack
from .element.core import RowLayout

from .element.display import VerticalSpace
from .element.display import Html
from .element.display import Text
from .element.display import Label
from .element.display import Image
from .element.display import Audio
from .element.display import Video
from .element.display import MatPlot
from .element.display import Hline
from .element.display import CodeBlock
from .element.display import ProgressBar
from .element.display import Alert
from .element.display import ButtonLabels
from .element.display import BarLabels
from .element.display import CountUp
from .element.display import CountDown

from .element.input import TextEntry
from .element.input import TextArea
from .element.input import RegEntry
from .element.input import PasswordEntry
from .element.input import NumberEntry
from .element.input import SingleChoice
from .element.input import MultipleChoice
from .element.input import SingleChoiceList
from .element.input import MultipleChoiceList
from .element.input import SingleChoiceButtons
from .element.input import SingleChoiceBar
from .element.input import MultipleChoiceButtons
from .element.input import MultipleChoiceBar
from .element.input import SelectPageList

from .element.action import SubmittingButtons
from .element.action import JumpButtons
from .element.action import DynamicJumpButtons
from .element.action import JumpList
from .element.action import Button

from .element.misc import Style
from .element.misc import HideNavigation
from .element.misc import JavaScript
from .element.misc import WebExitEnabler
from .element.misc import Value
from .element.misc import Data
from .element.misc import Callback
from .element.misc import RepeatedCallback

from .util import emoji
from .util import icon

from .util import is_element
from .util import is_input_element
from .util import is_label
from .util import is_page
from .util import is_section


# -*- coding: utf-8 -*-

"""
Das Modul definiert alle Exceptions des Frameworks
"""

class AlfredError(Exception):
    u"""
    Jede Exception des Frameworks ist von dieser Klasse abgeleitet.
    """
    pass

class ValidationError(AlfredError): pass

class AbortMove(AlfredError): pass

class MoveError(AlfredError):
    pass


class SavingAgentRunException(AlfredError):
    pass


class SavingAgentException(AlfredError):
    pass

class SessionTimeout(AlfredError): 
    pass


"""
Provides fundamental element classes.

.. moduleauthor: Johannes Brachem <jbrachem@posteo.de>

"""

from abc import ABC, abstractproperty, abstractmethod
from typing import List
from typing import Tuple
from typing import Union
from typing import Iterator
from dataclasses import dataclass

from jinja2 import Environment
from jinja2 import PackageLoader
from jinja2 import Template

from .. import alfredlog
from ..messages import MessageManager
from ..exceptions import AlfredError
from .._helper import fontsize_converter
from .._helper import check_name
from .._helper import inherit_kwargs


#: jinja Environment giving access to included jinja-templates.
jinja_env = Environment(loader=PackageLoader(__name__, "templates"))


class Element:
    """
    Element baseclass, providing basic functionality for all elements.

    Args:
        name: Name of the element. This should be a unique identifier.
            It will be used to identify the corresponding data in the
            final data set.
        font_size: Font size for text in the element. You can use a 
            keyword or an exact specification. The available keywords 
            are 'tiny', 'small', 'normal', 'big', and 'huge'. The exact
            specification shoul ideally include a unit, such as '1rem',
            or '12pt'. If you supply an integer without a unit, a unit
            of 'pt' will be assumed. Defaults to 'normal'.
        align: Horizontal alignment of text in the element. Does not
            usually apply to labels. Think of it as an alignment that
            applies to the innermost layer of an element (while labels
            are generally located at outer layers). See
            :class:`.LabelledElement` for more on labelled elements.
            Can be 'left' (default), 'center', 'right', or 'justify'.
        position: Horizontal position of the full element on the
            page. Values can be 'left', 'center' (default), 'end',
            or any valid value for the justify-content
            `flexbox utility <https://getbootstrap.com/docs/4.0/utilities/flex/#justify-content>`_.
            Takes effect only, when the element is not
            full-width.
        width: Defines the horizontal width of the element from
            small screens upwards. It is always full-width on extra
            small screens. Possible values are 'narrow', 'medium',
            'wide', and 'full'. For more detailed control, you can
            define the :attr:`.element_width` attribute.
        height: Vertical height of the elements display area. Supply a
            string with a unit, e.g. "80px". Usually, the default is
            fine. For adding vertical space to a page, you should prefer
            the :class:`.VerticalSpace` element, as it is sematically
            more clear.
        showif: A dictionary, defining conditions that must be met
            for the element to be shown. The conditions take the form of
            key-value pairs, where each key is an element name and the
            value is the required input. See :attr:`showif` for details.
        instance_log: If *True*, the element will use an
            instance-specific logger, thereby allowing detailed fine-
            tuning of its logging behavior.

    See Also:
        * How to create a custom element

    Notes:
        The Element does not have its own display. It is used only
        to inherit functionality.


    """

    #: Base template for the element, which will be used to hold the
    #: rendered element template. Gets rendered by :attr:`.Element.web_widget`
    base_template: Template = jinja_env.get_template("html/Element.html.j2")

    #: The element's specific, inner template. Gets rendered by
    #: :meth:`.render_inner_html`
    element_template: Template = None

    _inherited_kwargs = {}

    def __init__(
        self,
        name: str = None,
        font_size: Union[str, int] = None,
        align: str = "left",
        width: str = "full",
        height: str = None,
        position: str = "center",
        showif: dict = None,
        instance_log: bool = False,
    ):

        self.name: str = name  # documented in getter property
        self.page = None  # documented in getter
        self.exp = None  # documented in getter
        self.experiment = None  # documented in getter

        #: Alignment of inner element (does not apply to labels)
        self.align = align

        #: Vertical height of the element
        self.height = height

        self.font_size = font_size  # documented in getter property
        self.width = width  # documented in getter property
        self.position = position  # documented in getter property
        self._element_width = None  # documented in getter property

        # showifs and filters
        self.showif = showif if showif else {}  # documented in getter property

        #: Flag, indicating whether the element has a showif condition
        #: that includes an element on the same page
        self._showif_on_current_page = False
        self._should_be_shown = True  # documented in getter property

        self.display_standalone = True

        # additional code
        self._css_code = []  # documented in getter property
        self._css_urls = []  # documented in getter property
        self._js_code = []  # documented in getter property
        self._js_urls = []  # documented in getter property

        #: Boolean flag, indicating whether the element should spawn
        #: its own logger, or use the class-specific logger.
        #: Can be set to *True* to allow for very fine-grained logging.
        #: In most cases, it is fine to leave it at the default (*False*)
        self.instance_log: bool = instance_log

        #: A :class:`~.QueuedLoggingInterface`, offering logging
        #: through the methods *debug*, *info*, *warning*, *error*,
        #: *exception*, and *log*.
        self.log = alfredlog.QueuedLoggingInterface(base_logger=__name__)

        if position != "center" and width == "full":
            self.log.warning(
                (
                    "You have changed the value of 'position' on a full-width element. "
                    "That will most likely not have an effect. Did you mean to change 'align'?"
                )
            )

    @property
    def page(self):
        """
        alfred3.Page: The page to which this element belongs.
        """
        return self._page

    @page.setter
    def page(self, page):
        self._page = page

    @property
    def exp(self):
        """
        alfred3.ExperimentSession: The experiment session to which this
        element belongs.
        """
        return self._exp

    @exp.setter
    def exp(self, exp):
        self._exp = exp

    @property
    def experiment(self):
        """
        Alias for :attr:`.exp`
        """
        return self._exp

    @experiment.setter
    def experiment(self, exp):
        self.exp = exp

    @property
    def display_standalone(self):
        """
        bool: If *True* (default), the element will be displayed as 
        usual on its own. If *False*, the element will not be displayed
        unless you incorporate its :attr:`.Element.web_widget` in some
        other way.

        Notes:
            An element with ``display_standalone = False`` will still be validated.

        See Also:
            Similar to :attr:`.Element.should_be_shown`. The main 
            difference is that an element with 
            ``display_standalone = False`` will still be validated, while an
            element with ``should_be_shown = False`` will never be 
            validated.
        """
        return self._display_standalone
    
    @display_standalone.setter
    def display_standalone(self, value):
        self._display_standalone = value

    @property
    def showif(self) -> dict:
        """
        dict: Conditions that have to be met for the element to be shown.

        The showif dictionary can contain multiple conditions as key-
        value-pairs. The element will only be shown, if *all* conditions
        are met. You can use all names that show up in the main dataset.
        That includes:

        * The names of all input elements that were shown before the
          current page
        * The names of all input elements *on* the current page
        * The experiment metadata, including *exp_condition*,
          *exp_start_time*, and more. See :attr:`.Experiment.metadata`

        .. note::
            If you wish to implement more sophisticated conditions (e.g.
            linking conditions with 'or' instead of 'and'), you can
            do so by using if-statements in an *on_first_show* or
            *on_each_show* page-hook.

            Those conditions will not work for elements on the same page
            though. If you want to create complex showif conditions
            depending on elements on the same page, you have to
            implement them in JavaScript yourself. See :meth:`.add_js`
            and :class:`.JavaScript` for information on how to add
            JavaScript.

        Examples:

            This is a simple showif based on experiment condition. The
            text element in this example will only be shown to subjects
            in the condition "one"::

                import alfred3 as al
                exp = al.Experiment()

                @exp.setup
                def setup(exp):
                    exp.condition = al.random_condition("one", "two")

                @exp.member
                class MyPage(al.Page):

                    def on_exp_access(self):
                        self += al.Text("This is text", showif={"exp_condition": "one"})

            This is a more complex condition using the hook method. The
            text element on *page2* will only be show to subjects, if
            they spent more than 20 seconds on *page1*::

                import alfred3 as al
                exp = al.Experiment()
                exp += al.Page(name="FirstPage")

                @exp.member
                class SecondPage(al.Page):

                    def on_first_show(self):
                        if sum(self.exp.page1.durations) > 20:
                            self += al.Text("This is text")

        See Also:
            Page hooks

        Todo:
            * Put link to page hooks explanation here.

        """
        return self._showif

    @showif.setter
    def showif(self, value: dict):
        if not isinstance(value, dict):
            raise TypeError("Showif must be of type 'dict'.")
        self._showif = value

    @property
    def converted_width(self) -> List[str]:
        """
        list: List of bootstrap column widths at different screen sizes.

        Converted from :attr:`.width`.
        """
        if self.width == "narrow":
            return ["col-12", "col-sm-6", "col-md-3"]
        elif self.width == "medium":
            return ["col-12", "col-sm-9", "col-md-6"]
        elif self.width == "wide":
            return ["col-12", "col-md-9"]
        elif self.width == "full":
            return ["col-12"]

    @property
    def width(self) -> str:
        """str: Element width"""
        return self._width

    @width.setter
    def width(self, value):
        if value is None:
            self._width = None
        elif value not in ["narrow", "medium", "wide", "full"]:
            raise ValueError(f"'{value}' is not a valid width.")
        self._width = value

    @property
    def position(self) -> str:
        """
        str: Position of the whole element on the page.

        Determines the element position, when an element is *not*
        full-width.
        """
        return self._position

    @position.setter
    def position(self, value):
        if value == "left":
            self._position = "start"
        elif value == "right":
            self._position = "end"
        else:
            self._position = value

    @property
    def font_size(self) -> int:
        """int: Font size"""
        return self._font_size

    @font_size.setter
    def font_size(self, value):
        self._font_size = fontsize_converter(value)

    @property
    def element_width(self) -> str:
        """
        str: Returns a string of column width definitions.

        **Manually setting the width**

        The element width list can contain up to 5 width definitions,
        specified as integers from 1 to 12. The integers refer to the
        five breakbpoints in `Bootstrap 4's 12-column-grid system`_, i.e.
        [xs, sm, md, lg, xl]::

            >>> element = Element()
            >>> element.element_width = [12, 8, 8, 7, 6]
            element.element_width
            "col-12 col-sm-8 col-md-8 col-lg-7 col-xl-6"

        **Width resultion order**

        If there is no width defined for a certain screen size,
        the next smaller entry is used. For example, the following
        definition will lead to full width on extra small screens
        and 8/12 widths on all larger widths::

            element = Element()
            element.element_width = [12, 8]

        To make an element full-width on extra small and small
        screens and half-width on medium, large and extra large
        screens, follow this example::

            element = Element()
            element.element_width = [12, 12, 6]

        .. _Bootstrap 4's 12-column-grid system: https://getbootstrap.com/docs/4.0/layout/grid/

        """
        if self.width is not None:
            return " ".join(self.converted_width)

        width = self._element_width if self._element_width is not None else ["col-12"]
        if self.experiment.config.getboolean("layout", "responsive", fallback=True):
            return " ".join(width)
        else:
            return width[0]

    @element_width.setter
    def element_width(self, value: List[int]):
        try:
            for v in value:
                if not isinstance(v, int):
                    raise TypeError("Element width must be set as a list of integers.")
        except TypeError:
            raise TypeError("Element width must be set as a list of integers.")
        self._element_width = value

    @property
    def name(self) -> str:
        """
        str: Unique identifier for the element.

        The element name will be used to identify the corresponding
        input in the final dataset.
        """
        return self._name

    @name.setter
    def name(self, name):
        if name is None:
            self._name = None

        elif name is not None:
            check_name(name)

        self._name = name

    @property
    def should_be_shown(self) -> bool:
        """
        bool: Boolean, indicating whether the element is meant to be shown.

        Evaluates all *showif* conditions. Can be set manually.
        Returns *False*, if the element's parent
        page should not be shown.
        """
        cond1 = self._should_be_shown
        cond2 = all(self._evaluate_showif())
        cond3 = self.page.should_be_shown
        return cond1 and cond2 and cond3

    @should_be_shown.setter
    def should_be_shown(self, b: bool):
        if not isinstance(b, bool):
            raise TypeError("should_be_shown must be an instance of bool")
        self._should_be_shown = b

    @property
    def section(self):
        """The direct parent section of this element's page."""
        return self.page.section

    @property
    def tree(self) -> str:
        """
        str: String, giving the exact position in the experiment.

        The tree is composed of the names of all sections
        and the element's page dots. Example for an element that belongs
        to a page "hello_world" that was added directly to the experiment::

            _root._content.hello_world

        ``_root`` and ``_content`` are basic sections that alfred always
        includes in an experiment.

        """
        return self.page.tree

    @property
    def short_tree(self) -> str:
        """
        str: String, giving the exact position in the experiment.

        This version of the tree omits the ``_root`` part.
        """

        return self.tree.replace("_root.", "")

    @property
    def css_code(self) -> List[Tuple[int, str]]:
        """List[tuple]: A list of tuples, which contain a priority and CSS code."""
        return self._css_code

    @property
    def css_urls(self) -> List[Tuple[int, str]]:
        """List[tuple]: A list of tuples, which contain a priority and an url pointing
        to CSS code."""
        return self._css_urls

    @property
    def js_code(self) -> List[Tuple[int, str]]:
        """List[tuple]: A list of tuples, which contain a priority and Javascript."""
        return self._js_code

    @property
    def js_urls(self) -> List[Tuple[int, str]]:
        """List[tuple]: A list of tuples, which contain a priority and an url pointing
        to JavaScript."""
        return self._js_urls

    @property
    def template_data(self) -> dict:
        """
        dict: Dictionary of data to be passed on to jinja templates.

        When deriving a new element class, you will often want to
        redefine this property to add template data. When doing so,
        remember to retrieve the basic template data with ``super()``::

            import alfred3 as al

            class NewElement(al.Element):

                @property
                def template_data(self):
                    d = super().template_data
                    d["my_value"] = "this is my value"

                    return d    # don't forget to return the dictionary!

        The call ``super().template_data`` applies the parent classes
        code to the current object. That way, you only need to define
        values that differ from the parent class.

        .. note::
            Be aware that, by default, the same template data will be
            passed to :attr:`.element_template` and :attr:`.base_template`.

        """
        d = {}
        d["css_class_element"] = self.css_class_element
        d["css_class_container"] = self.css_class_container
        d["name"] = self.name
        d["position"] = self.position
        d["element_width"] = self.element_width
        d["hide"] = "hide" if self._showif_on_current_page is True else ""
        d["align"] = f"text-{self.align}"
        d["align_raw"] = self.align
        d["fontsize"] = f"font-size: {self.font_size};" if self.font_size is not None else ""
        d["height"] = f"height: {self.height};" if self.height is not None else ""
        d["responsive"] = self.experiment.config.getboolean("layout", "responsive")
        return d

    # Private methods start here ---------------------------------------

    def _evaluate_showif(self) -> List[bool]:
        """Checks the showif conditions that refer to previous pages.

        Returns:
            list: A list of booleans, indicating for each condition,
            whether it is met or not.
        """

        if self.showif:
            conditions = []
            for name, condition in self.showif.items():

                # skip current page (showifs for current pages are checked elsewhere)
                if name in self.page.all_input_elements:
                    continue

                val = self.exp.data_manager.flat_session_data[name]
                conditions.append(condition == val)

            return conditions
        else:
            return [True]

    def _activate_showif_on_current_page(self):
        """Adds JavaScript to self for dynamic showif functionality."""
        pg = self.page.all_input_elements
        on_current_page = dict([cond for cond in self.showif.items() if cond[0] in pg])

        if on_current_page:

            t = jinja_env.get_template("js/showif.js.j2")
            js = t.render(showif=on_current_page, element=self.name)
            self.js_code.append((7, js))
            self._showif_on_current_page = True

    # Public methods start here ----------------------------------------

    def added_to_experiment(self, experiment):
        """
        Tells the element that it was added to an experiment.

        The experiment is made available to the element, and the
        element's logging interface initializes its experiment-specific
        logging.

        This is also the place where the element's name is checked for
        experiment-wide uniqueness.

        Args:
            experiment: The alfred experiment to which the element was
                added.

        """

        if self.name in experiment.root_section.all_updated_elements:
            raise AlfredError(f"Element name '{self.name}' is already present in the experiment.")

        if self.name in experiment.data_manager.flat_session_data:
            raise AlfredError(f"Element name '{self.name}' conflicts with a protected name.")

        self.experiment = experiment
        self.exp = experiment
        self.log.add_queue_logger(self, __name__)

    def added_to_page(self, page):
        """
        Tells the element that it was added to a page.

        The page and the experiment are made available to the element.

        Args:
            page: The page to which the element was added.

        """
        from .. import page as pg

        if not isinstance(page, pg._PageCore):
            raise TypeError()

        self.page = page
        if self.name is None:
            self.name = self.page._generate_element_name(self)

        if self.page.experiment and not self.experiment:
            self.added_to_experiment(self.page.experiment)

    def _prepare_web_widget(self):
        """
        Wraps :meth:`.prepare_web_widget` to allow for additional, generic
        preparations that are the same for all elements.

        This is useful, because :meth:`.prepare_web_widget` is often
        redefined in derived elements.
        """
        self._activate_showif_on_current_page()
        self.prepare_web_widget()

    def prepare_web_widget(self):
        """
        Hook for computations for preparing an element's web widget.

        This method is supposed to be overridden by derived elements if
        necessary.
        """
        pass

    # Magic methods start here -----------------------------------------

    def __str__(self):
        return f"{type(self).__name__}(name='{self.name}')"

    def __repr__(self):
        return self.__str__()

    # abstract attributes start here -----------------------------------

    @property
    def _inner_html(self) -> str:
        """
        Shortcut for rendering the element template on its own with
        the template data.

        Notes:
            This should not be used together with :attr:`.web_widget`,
            because it will lead to :attr:`.template_data` being called
            twice. Calling the template data twice will empty all
            corrective hints, causing them to not be displayed at all.

        """
        return self.render_inner_html(self.template_data)

    def render_inner_html(self, template_data: dict) -> str:
        """
        Renders the element template :attr:`~.element_template`.

        Args:
            template_data: A dictionary of data for rendering the
                template.

        Notes:
            If no `element_template` is defined, `None` is returned. Usually,
            the inner html gets placed into the higher-level
            :attr: `.base_template`, when :attr:`Element.web_widget` gets called.

        Returns:
            str: Inner html code for this element.
        """
        if self.element_template is not None:
            return self.element_template.render(template_data)
        else:
            return None

    @property
    def web_widget(self) -> str:
        """
        The element's rendered html code for display on a page.

        Notes:
            This function gets the :attr:`.template_data`, and uses it
            to first call :meth:`.render_inner_html`, and then render
            the :attr:`.base_template`.

        Returns:
            str: The full html code for this element.
        """
        d = self.template_data
        d["html"] = self.render_inner_html(d)
        return self.base_template.render(d)

    @property
    def css_class_element(self) -> str:
        """
        Returns the name of the element's CSS class.

        On the webpage generated by alfred3, all elements reside in some
        sort of container which has a CSS class of the element's
        *css_class_element*. The name is composed as the element's class
        name, followed by ``-element``.

        Examples:

            This is a simplified illustration of the html structure of
            a text element. Let's say, the element was instantiated
            with the name "example_text". The element's *css_class_element*
            will be *Text-element*:

                >>> import alfred3 as al
                >>> ex = Text("This is an example", name="example_text")
                >>> ex.css_class_element
                Text-element

            The html code generated by this element is structured as
            follows

            .. code-block:: html

                <div class="Text-element-container" id="example_text-container">
                  ...
                  <div class="Text-element" id="example_text">
                    <p>This is an example</p>
                  </div>
                  ...
                </div>

            Note that the element receives two CSS classes and two IDs,
            one each for the outer container and one for the innermost
            layer. The outer container includes the element's base
            template (e.g. *Element.html*, or *LabelledElement.html*),
            while the inner layer includes the specific element-
            template (e.g. *TextElement.html*).

        """
        return f"{type(self).__name__}-element"

    @property
    def css_class_container(self) -> str:
        """
        str: Returns the name the element container's CSS class.

        The class can be used for CSS styling.

        See Also:
            See :attr:`.css_class_element` for details.
        """
        return f"{self.css_class_element}-container"

    def add_css(self, code: str, priority: int = 10):
        """
        Adds CSS to the element.

        This is most useful when writing new elements. To simply add
        element-related CSS code to a page, usually the :class:`.CSS`
        element is a better choice.

        Args:
            code: Css code
            priority: Can be used to influence the order in which code
                is added to the page. Sorting is ascending, i.e. the
                lowest numbers appear closest to the top of the page.

        See Also:
            * The :class:`.Style` element can be used to add generic CSS
              code to a page.
            * See :attr:`.Element.css_class_element` and
              :attr:`.Element.css_class_container` for information on
              element CSS classes and IDs.

        """
        self._css_code.append((priority, code))

    def add_js(self, code: str, priority: int = 10):
        """
        Adds Javascript to the element.

        This is most useful when writing new elements. To simply add
        element-related JavaScript code to a page, usually the
        :class:`.JavaScript` element is a better choice.

        Args:
            code: Css code
            priority: Can be used to influence the order in which code
                is added to the page. Sorting is ascending, i.e. the
                lowest numbers appear closest to the top of the page.

        See Also:
            * The :class:`.JavaScript` element can be used to add generic
              JavsScript code to a page.
            * See :attr:`.Element.css_class_element` and
              :attr:`.Element.css_class_container` for information on
              element CSS classes and IDs.

        """
        self._js_code.append((priority, code))


class RowLayout:
    """
    Provides layouting functionality for responsive horizontal
    positioning of elements.

    Default behavior is to have equal-width columns with an automatic
    breakpoint on extra small screens (i.e. all columns get the bootstrap
    class 'col-sm' by default).

    The layout's width attributes can be accessed and changed to customize
    appearance. In this example, we change the width of the columns on
    screens of "small" and bigger width, so that we have narrow columns
    to the right and left (each taking up 2/12 of the available space),
    and one wide column (taking up 8/12 of the space) in the middle. On
    "extra small" screens, the columns will be stacked vertically and
    each take up the full width.

    You can define widths for five breakpoints individually, allowing
    for fine-grained control (see attributes).

    Args:
        ncols: Number of columns to arrange in a row.
        valign_cols: List of vertical column alignments. Valid values
            are 'auto' (default), 'top', 'center', and 'bottom'.
        responsive: Boolean, indicating whether breakpoints should
            be responsive, or not.

    Examples:

        ::

            layout = RowLayout(ncols=3) # 3 columns of equal width
            layout.width_sm = [2, 8, 2]

    """

    def __init__(self, ncols: int, valign_cols: List[str] = None, responsive: bool = True):
        """Constructor method."""
        self.ncols: int = ncols
        self._valign_cols = valign_cols if valign_cols is not None else []
        self.responsive: bool = responsive  # documented in getter

        self._width_xs: List[int] = None  # documented in getter
        self._width_sm: List[int] = None  # documented in getter
        self._width_md: List[int] = None  # documented in getter
        self._width_lg: List[int] = None  # documented in getter
        self._width_xl: List[int] = None  # documented in getter

    @property
    def ncols(self):
        """
        int: Number of columns
        """
        return self._ncols

    @ncols.setter
    def ncols(self, value):
        self._ncols = value

    @property
    def responsive(self):
        """
        bool: Indicates whether breakpoints should be responsive, or not.
        """
        return self._responsive

    @responsive.setter
    def responsive(self, value):
        self._responsive = value

    @property
    def width_xs(self):
        """
        List[int]: List of column widths on screens of size 'xs' or bigger
        (<576px). Content must be integers between 1 and 12.
        """
        return self._width_xs

    @width_xs.setter
    def width_xs(self, value):
        try:
            assert isinstance(value[0], int)
        except (AssertionError, TypeError):
            raise ValueError("width must be a list of integers")
        
        try:
            assert len(value) <= self.ncols
        except AssertionError:
            raise ValueError(f"Number of widths must be smaller or equal to the number of columns ({self.ncols}), not {len(value)}.")

        self._width_xs = value

    @property
    def width_sm(self):
        """
        List[int]: List of column widths on screens of size 'sm' or bigger
        (>=576px). Content must be integers between 1 and 12.
        """
        return self._width_sm

    @width_sm.setter
    def width_sm(self, value):
        try:
            assert isinstance(value[0], int)
        except AssertionError:
            raise ValueError("width must be a list of integers")

        try:
            assert len(value) <= self.ncols
        except AssertionError:
            raise ValueError(f"Number of widths must be smaller or equal to the number of columns ({self.ncols}), not {len(value)}.")

        self._width_sm = value

    @property
    def width_md(self):
        """
        List[int]: List of column widths on screens of size 'md' or bigger
        (>=768px). Content must be integers between 1 and 12.
        """
        return self._width_md

    @width_md.setter
    def width_md(self, value):
        try:
            assert isinstance(value[0], int)
        except AssertionError:
            raise ValueError("width must be a list of integers")
        
        try:
            assert len(value) <= self.ncols
        except AssertionError:
            raise ValueError(f"Number of widths must be smaller or equal to the number of columns ({self.ncols}), not {len(value)}.")

        self._width_md = value

    @property
    def width_lg(self):
        """
        List[int]: List of column widths on screens of size 'lg' or bigger
        (>=992px). Content must be integers between 1 and 12.
        """
        return self._width_lg

    @width_lg.setter
    def width_lg(self, value):
        try:
            assert isinstance(value[0], int)
        except AssertionError:
            raise ValueError("width must be a list of integers")
        
        try:
            assert len(value) <= self.ncols
        except AssertionError:
            raise ValueError(f"Number of widths must be smaller or equal to the number of columns ({self.ncols}), not {len(value)}.")

        self._width_lg = value

    @property
    def width_xl(self):
        """
        List[int]: List of column widths on screens of size 'xl' or bigger
        (>=1200px). Content must be integers between 1 and 12.
        """
        return self._width_xl

    @width_xl.setter
    def width_xl(self, value):
        try:
            assert isinstance(value[0], int)
        except AssertionError:
            raise ValueError("width must be a list of integers")
        
        try:
            assert len(value) <= self.ncols
        except AssertionError:
            raise ValueError(f"Number of widths must be smaller or equal to the number of columns ({self.ncols}), not {len(value)}.")

        self._width_xl = value

    def col_breaks(self, col: int) -> str:
        """
        Returns the column breakpoints for a specific column as
        strings for use as bootstrap classes.

        Args:
            col: Column index (starts at 0)

        Returns:
            str: Column breakpoints for a specific column
        """
        xs = self.format_breaks(self.width_xs, "xs")[col]
        sm = self.format_breaks(self.width_sm, "sm")[col]
        md = self.format_breaks(self.width_md, "md")[col]
        lg = self.format_breaks(self.width_lg, "lg")[col]
        xl = self.format_breaks(self.width_xl, "xl")[col]

        if self.responsive:
            breaks = [xs, sm, md, lg, xl]
            if breaks == ["", "", "", "", ""]:
                return "col-sm"
            else:
                return " ".join(breaks)
        else:  # set breaks to a fixed value
            breaks = self.format_breaks(self.width_sm, "xs")[col]  # this is ONE value
            out = breaks if breaks != "" else "col"
            return out

    def format_breaks(self, breaks: List[int], bp: str) -> List[str]:
        """
        Takes a list of column sizes (in integers from 1 to 12) and
        returns a corresponding list of formatted Bootstrap column
        classes.

        Args:
            breaks: List of integers, indicating the breakpoints.
            bp: Specifies the relevant bootstrap breakpoint. (xs, sm,
                md, lg, or xl).

        Returns:
            List[str]: List of ready-to-use bootstrap column classes
        """
        try:
            if len(breaks) > self.ncols:
                raise ValueError("Break list length must be <= number of elements.")
        except TypeError:
            pass

        out = []
        for i in range(self.ncols):
            try:
                n = breaks[i]
            except (IndexError, TypeError):
                out.append("")
                continue

            if not isinstance(n, int):
                raise TypeError("Break values must be of type integer.")
            if not n >= 1 and n <= 12:
                raise ValueError("Break values must be between 1 and 12.")

            if bp == "xs":
                out.append(f"col-{n}")
            else:
                out.append(f"col-{bp}-{n}")

        return out

    @property
    def valign_cols(self) -> List[str]:
        """
        List[str]: Vertical column alignments.

        Valid values are 'auto' (default), 'top', 'center', and 'bottom'.
        Can be specified upon initalization or modified as instance
        attribute.

        Each element of the list refers to one column. If it contains
        fewer elements than the number of columns, the last entry of
        the list will be used as value for the unreferenced columns.

        Examples:

            All columns of the following layout will be aligned to the
            bottom of the row (specified upon initialization)::

                layout1 = RowLayout(ncols=3, valign_cols=["bottom"])

            The first column of the following layout will be aligned top,
            the 2nd and 3rd columns will be aligned bottom (specified after
            initialization)::

                layout2 = RowLayout(ncols=3)
                layout2.valign_cols = ["top", "bottom"]

        """

        out = []
        for i in range(self.ncols):
            try:
                n = self._valign_cols[i]
            except IndexError:
                try:
                    n = self._valign_cols[i - 1]
                except IndexError:
                    n = "center"

            if not isinstance(n, str):
                raise TypeError("Col position must be of type str.")

            if n == "auto":
                out.append("")
            elif n == "top":
                out.append("align-self-start")
            elif n == "center":
                out.append("align-self-center")
            elif n == "bottom":
                out.append("align-self-end")
            else:
                raise ValueError("Valign allowed values: 'auto', 'top', 'center', and 'bottom'.")

        return out

    @valign_cols.setter
    def valign_cols(self, value: List[str]):
        if len(value) > self.ncols:
            raise ValueError("Col position list length must be <= number of elements.")
        self._valign_cols = value


@dataclass
class _RowCol:
    """
    Just a little helper for handling columns inside a Row.

    :meta private:
    """

    breaks: str
    vertical_position: str
    element: Element
    id: str


@inherit_kwargs
class Row(Element):
    """
    Allows you to arrange up to 12 elements in a row.

    The row will arrange your elements using Bootstrap 4's grid system
    and breakpoints, making the arrangement responsive to different
    screen sizes. You can customize the behavior of the row for five
    different screen sizes (Bootstrap 4's default break points) with
    the width attributes of its layout attribute.

    If you don't specify breakpoints manually via the *layout* attribute,
    the columns will default to equal width and wrap on breakpoints
    automatically.

    Args:
        elements: The elements that you want to arrange in a row.
        valign_cols: List of vertical column alignments. Valid values
            are 'auto' (default), 'top', 'center', and 'bottom'. The
            elements of the list correspond to the row's columns. See
            :attr:`.RowLayout.valign_cols`
        elements_full_width: A switch, telling the row whether you wish
            it to resize all elements in it to full-width (default: True).
            This switch exists, because some elements might default to
            a smaller width, but when using them in a Row, you usually
            want them to span the full width of their column.

        {kwargs}

    Notes:

        In Bootstrap's grid, the horizontal space is divided into 12
        equally wide units. You can define the horizontal width of a
        column by assigning it a number of those units. A column of
        width 12 will take up all available horizontal space, other
        columns will be placed below such a full-width column.

        You can define the column width for each of five breakpoints
        separately. The definition will be valid for screens of the
        respective size up to the next breakpoint.

        See https://getbootstrap.com/docs/4.5/layout/grid/#grid-options
        for detailed documentation of how Bootstrap's breakpoints work.

    Examples:

        A minimal experiment with a row::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class HelloWorld(al.Page):
                name = "hello_world"

                def on_exp_access(self):
                    el1 = al.Text("text")
                    el2 = al.TextEntry(toplab="lab", name="example")

                    self += al.Row(el1, el2)

        The arrangement will look like this::

            |======================|=====================|
            |   el1                |  el2                |
            |======================|=====================|

        An example with customized widths, where we redefine the
        "sm" breakpoint::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class HelloWorld(al.Page):
                name = "hello_world"

                def on_exp_access(self):
                    el1 = al.Text("text")
                    el2 = al.TextEntry(toplab="lab", name="example")

                    row = al.Row(el1, el2)
                    row.layout.width_sm = [2, 10]

                    self += row

        The arrangement will look like this::

            |===========|===============================|
            |   el1     |  el2                          |
            |===========|===============================|


    """

    element_template = jinja_env.get_template("html/Row.html.j2")

    def __init__(
        self,
        *elements: Element,
        valign_cols: List[str] = None,
        elements_full_width: bool = True,
        height: str = "auto",
        name: str = None,
        showif: dict = None,
        **kwargs,
    ):
        """Constructor method."""
        super().__init__(name=name, showif=showif, height=height, **kwargs)

        self.elements: list = elements  # documented in getter
        self.layout = RowLayout(
            ncols=len(self.elements), valign_cols=valign_cols
        )  # documented in getter
        self.elements_full_width: bool = elements_full_width  # documented in getter

    @property
    def elements(self):
        """
        List[Element]: List of the elements in this row.
        """
        return self._elements

    @elements.setter
    def elements(self, value):
        self._elements = value

    @property
    def layout(self):
        """
        RowLayout: The layout of this row, can be used to finetune
        column widths via its attributes. See :class:`.RowLayout` for
        the available width attributes.
        """
        return self._layout

    @layout.setter
    def layout(self, value):
        self._layout = value

    @property
    def elements_full_width(self):
        """
        bool: If *True*, all elements will take up the full horizontal
        space of their column, regardless of the element's
        :attr:`.Element.element_width` and :attr:`.Element.width`
        attributes.

        This is done by default for better default layouting.
        """
        return self._elements_full_width

    @elements_full_width.setter
    def elements_full_width(self, value):
        self._elements_full_width = value

    def added_to_page(self, page):

        super().added_to_page(page)

        for element in self.elements:
            if element is None:
                continue
            element.display_standalone = False
            page += element

            if self.elements_full_width:
                element.width = "full"

    def _prepare_web_widget(self):

        for element in self.elements:
            element.prepare_web_widget()

        super()._prepare_web_widget()

    @property
    def _cols(self) -> Iterator:
        """Yields preprocessed ``_RowCol`` columns ."""
        for i, element in enumerate(self.elements):
            col = _RowCol(
                breaks=self.layout.col_breaks(col=i),
                vertical_position=self.layout.valign_cols[i],
                element=element,
                id=f"{self.name}_col{i+1}",
            )
            yield col

    @property
    def template_data(self):

        d = super().template_data
        d["columns"] = self._cols
        d["name"] = self.name
        return d


@inherit_kwargs
class Stack(Row):
    """
    Stacks multiple elements on top of each other.

    Stacks are intended for use in Rows. They allow you to flexibly
    arrange elements in a grid.

    Args:
        *elements: The elements to stack.

        {kwargs}

    Examples:

        A minimal experiment with a stack in a row::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class HelloWorld(al.Page):
                name = "hello_world"

                def on_exp_access(self):
                    el1 = al.Text("text")
                    el2 = al.TextEntry(toplab="lab", name="example")
                    el3 = al.Text("long text")

                    stack = al.Stack(el1, el2)
                    self += al.Row(stack, el3)

        The arrangement will look like this::

            |=========|========|
            |   el1   |        |
            |=========|  el3   |
            |   el2   |        |
            |=========|========|

    """

    def __init__(self, *elements: Element, **kwargs):
        """Constructor method."""
        super().__init__(*elements, **kwargs)
        self.layout.width_xs = [12 for element in elements]


@inherit_kwargs
class LabelledElement(Element):
    """
    An intermediate Element class which provides support for labels.

    This class is used as a base class for all elements that come
    equipped with labels.

    Args:
        toplab, bottomlab, leftlab, rightlab: Labels above, below, left
            and right of the element.

        layout: Can be one of the following: 1) An instance of
            :class:`.RowLayout`, or 2) a tuple of integers, specifying 
            the allocation of horizontal space between leftlab, main 
            element widget and rightlab on small screens upwards. 

            Option 1) offers fine-tuned flexibility, 2) uses a default
            RowLayout and changes the :attr:`.RowLayout.width_sm` 
            attribute.

            By default, the layout is set automatically depending on the
            specification of the left and right labels.


        {kwargs}

    Notes:
        The labelled element is not supposed to be included on a page on
        its own - it is meant for easy derivation of new elements.

    """

    base_template = jinja_env.get_template("html/LabelledElement.html.j2")

    def __init__(
        self,
        toplab: str = None,
        leftlab: str = None,
        rightlab: str = None,
        bottomlab: str = None,
        layout: Union[RowLayout, Tuple[int]] = None,
        **kwargs,
    ):
        """Constructor method."""
        super().__init__(**kwargs)

        self._ncols = len([x for x in [leftlab, rightlab, 1] if x is not None])
        self._input_col = 1 if leftlab is not None else 0

        if isinstance(layout, RowLayout):
            self.layout = layout
        else:
            self.layout = RowLayout(ncols=self._ncols)
            self.layout.valign_cols = ["center"]

            if leftlab and rightlab:
                width_sm = layout if layout is not None else [2, 8, 2]
            elif leftlab:
                width_sm = layout if layout is not None else [3, 9]
            elif rightlab:
                width_sm = layout if layout is not None else [9, 3]
            else:
                width_sm = [12]

            self.layout.width_sm = width_sm

        self.toplab = toplab
        self.leftlab = leftlab
        self.rightlab = rightlab
        self.bottomlab = bottomlab

    @property
    def layout(self):
        """
        RowLayout: Controls the allocation of horizontal space between
        the left and right label, as well as the main element.
        """
        return self._layout

    @layout.setter
    def layout(self, value: RowLayout):
        try:
            if not value.ncols == self._ncols:
                raise AlfredError(
                    (
                        "The number of layout columns must match the specification of "
                        f"left and right labels. In this case, you need {self._ncols} columns."
                    )
                )
            self._layout = value
        except AttributeError:
            raise TypeError("Layout must be an instance of 'alfred3.RowLayout'.")


    def added_to_page(self, page):

        super().added_to_page(page)

        for lab in ["toplab", "leftlab", "rightlab", "bottomlab"]:
            if getattr(self, lab):
                getattr(self, lab).name = f"{self.name}_{lab}"

    def added_to_experiment(self, experiment):

        super().added_to_experiment(experiment)
        self._layout.responsive = self.experiment.config.getboolean("layout", "responsive")

        if self.toplab:
            self.toplab.added_to_experiment(experiment)

        if self.leftlab:
            self.leftlab.added_to_experiment(experiment)

        if self.rightlab:
            self.rightlab.added_to_experiment(experiment)

        if self.bottomlab:
            self.bottomlab.added_to_experiment(experiment)

    @property
    def toplab(self):
        """.Label: Label above of the main element widget."""
        return self._toplab

    @toplab.setter
    def toplab(self, value: str):
        from .display import Label

        if isinstance(value, Label):
            self._toplab = value
        elif isinstance(value, str):
            self._toplab = Label(text=value, align="center", name=f"{self.name}_toplab")
        else:
            self._toplab = None

    @property
    def bottomlab(self):
        """.Label: Label below of the main element widget."""
        return self._bottomlab

    @bottomlab.setter
    def bottomlab(self, value: str):
        from .display import Label

        if isinstance(value, Label):
            self._bottomlab = value
        elif isinstance(value, str):
            self._bottomlab = Label(text=value, align="center", name=f"{self.name}_bottomlab")
        else:
            self._bottomlab = None

    @property
    def leftlab(self):
        """.Label: Label to the left of the main element widget."""
        return self._leftlab

    @leftlab.setter
    def leftlab(self, value: str):
        from .display import Label

        if isinstance(value, Label):
            self._leftlab = value
            self._leftlab.layout = self._layout
            self._leftlab.layout_col = 0
        elif isinstance(value, str):
            self._leftlab = Label(text=value, align="right", name=f"{self.name}_leftlab")
            self._leftlab.layout = self._layout
            self._leftlab.layout_col = 0
        else:
            self._leftlab = None

    @property
    def rightlab(self):
        """.Label: Label to the right of the main element widget."""
        return self._rightlab

    @rightlab.setter
    def rightlab(self, value: str):
        from .display import Label

        if isinstance(value, Label):
            self._rightlab = value
            self._rightlab.layout = self._layout
            self._rightlab.layout_col = self._input_col + 1
        elif isinstance(value, str):
            self._rightlab = Label(text=value, align="left", name=f"{self.name}_rightlab")
            self._rightlab.layout = self._layout
            self._rightlab.layout_col = self._input_col + 1
        else:
            self._rightlab = None

    @property
    def labels(self) -> str:
        """str: Returns the labels in a single, nicely formatted string."""

        labels = []
        if self.toplab:
            labels.append(f"toplab: '{self.toplab.text}'")
        if self.leftlab:
            labels.append(f"leftlab: '{self.leftlab.text}'")
        if self.rightlab:
            labels.append(f"rightlab: '{self.rightlab.text}'")
        if self.bottomlab:
            labels.append(f"bottomlab: '{self.bottomlab.text}'")

        if labels:
            return ", ".join(labels)

    @property
    def template_data(self):

        d = super().template_data
        d["toplab"] = self.toplab
        d["leftlab"] = self.leftlab
        d["rightlab"] = self.rightlab
        d["bottomlab"] = self.bottomlab
        d["input_breaks"] = self.layout.col_breaks(col=self._input_col)
        d["input_valign"] = self.layout.valign_cols[self._input_col]
        return d


@inherit_kwargs
class InputElement(LabelledElement):
    """
    Base class for elements that allow data input.

    Args:
        force_input: If `True`, users can  only progress to the next page
            if they enter data into this field. Note that a
            :class:`.NoValidationSection` or similar sections might
            overrule this setting.

            A general, experiment-wide setting for force_input can be
            placed in the config.conf (section "general"). That setting
            is used by default and can be overruled here for individual
            elements. Defaults to False.

            The experiment-wide default can be changed in config.conf.

        default: Default value. Type depends on the element type.
        prefix: Prefix for the input field.
        suffix: Suffix for the input field.
        description: An additional description of the element. This will
            show up in the alfred-generated codebook. It has
            no effect on the display of the experiment, as it only
            serves as a descriptor for humans.
        no_input_hint: Hint to be displayed if
            *force_input* set to True and no user input registered.
            Defaults to the experiment-wide default value
            specified in config.conf.

        {kwargs}

    Notes:
        The InputElement does not have its own display. It is used only
        to inherit functionality.

    """

    def __init__(
        self,
        toplab: str = None,
        force_input: bool = None,
        default: Union[str, int, float] = None,
        prefix: Union[str, Element] = None,
        suffix: Union[str, Element] = None,
        description: str = None,
        disabled: bool = False,
        no_input_hint: str = None,
        **kwargs,
    ):
        super().__init__(toplab=toplab, **kwargs)

        self.description = description  # documented in getter
        self.input = ""  # documented in getter
        self._force_input = force_input  # documented in getter property
        self._no_input_hint = no_input_hint
        self._default = default  # documented in getter property
        self._prefix = prefix  # documented in getter property
        self._suffix = suffix  # documented in getter property
        self.show_hints: bool = True
        self._hint_manager = MessageManager(default_level="danger")  # documented in getter
        self.disabled: bool = disabled  # documented in getter

        if default is not None:
            self.input = default

        if self._force_input and (self._showif_on_current_page or self.showif):
            raise ValueError(f"Elements with 'showif's can't be 'force_input' ({self}).")

    @property
    def show_hints(self):
        """
        bool: Flag, indicating whether corrective hints regarding
        this element should be shown.
        """
        return self._show_hints

    @show_hints.setter
    def show_hints(self, value):
        self._show_hints = value

    @property
    def description(self):
        """
        str: Detailed description of this element to be added to the
        automatically generated codebook
        """
        return self._description

    @description.setter
    def description(self, value):
        self._description = value

    @property
    def disabled(self):
        """
        bool: A boolean flag, indicating whether the element is disabled
        A disabled input element is shown and displays its input
        value, but subjects cannot enter any data.
        """
        return self._disabled

    @disabled.setter
    def disabled(self, value):
        self._disabled = value

    @property
    def hint_manager(self):
        """
        MessageManager: A :class:`.MessageManager`, handling the corrective hints
        for this element.
        """
        return self._hint_manager

    def _prepare_web_widget(self):

        super()._prepare_web_widget()

        try:
            self._prefix._prepare_web_widget()
        except AttributeError:
            pass

        try:
            self._suffix._prepare_web_widget()
        except AttributeError:
            pass

    @property
    def corrective_hints(self) -> Iterator[str]:
        """
        Shortcut for accessing the element's corrective hints.

        Yields:
            str: Corrective hint.
        """
        return self.hint_manager.get_messages()

    @property
    def debug_value(self) -> Union[str, None]:
        """
        Union[str, None]: Value to be used as a default in debug mode.

        This value is read from config.conf. Only used, if there is no
        dedicated default for this element.

        Notes:
            The property searches for an option of the form
            ``<element_class>_debug`` in the section *debug* of
            config.conf. If no option is found, the return value is
            *None*
        """
        name = f"{type(self).__name__}_default"
        return self.experiment.config.get("debug", name, fallback=None)

    @property
    def debug_enabled(self) -> bool:
        """
        bool: Boolean flag, indicating whether debug mode is enabled and
        default values should be set.
        """
        if self.experiment.config.getboolean("general", "debug"):
            if self.experiment.config.getboolean("debug", "set_default_values"):
                return True
        return False

    @property
    def default(self) -> Union[str, int, float]:
        """
        Union[str, int, float]: Default value of this element.

        The data type can vary, depending on the element.
        """
        if self._default is not None:
            return self._default
        elif self.debug_enabled:
            return self.debug_value
        else:
            return None

    @default.setter
    def default(self, value):
        self._default = value

    @property
    def force_input(self) -> bool:
        """
        bool: If *True*, subjects *must* fill this element to proceed.

        A general, experiment-wide setting for force_input can be placed
        in the config.conf (section "general").
        """
        if self._force_input is None:
            return self.exp.config.getboolean("general", "force_input")
        else:
            return self._force_input

    @force_input.setter
    def force_input(self, value: bool):
        if not isinstance(value, bool):
            raise ValueError("Force input must be a boolean value.")

        self._force_input = value

    @property
    def template_data(self) -> dict:

        d = super().template_data
        d["default"] = self.default
        d["prefix"] = self.prefix
        d["suffix"] = self.suffix
        d["input"] = self.input
        d["disabled"] = self.disabled
        if self.show_hints:
            d["corrective_hints"] = list(self.corrective_hints)
        return d

    def validate_data(self) -> bool:
        """
        Method for validation of input to the element.

        Returns:
            bool: *True*, if the input is correct and subjects may
            proceed to the next page, *False*, if the input is not
            in the correct form.
        """
        if not self.should_be_shown:
            return True

        elif self.force_input and not self.input:
            self.hint_manager.post_message(self.no_input_hint)
            return False

        else:
            return True

    @property
    def no_input_hint(self) -> str:
        """
        str: Hint for subjects, if they left a *force_input* field empty.
        """
        if self._no_input_hint:
            return self._no_input_hint
        return self.default_no_input_hint

    @property
    def default_no_input_hint(self) -> str:
        """
        str: Default hint if subject input is missing in *force_entry* elements.

        This value is read from config.conf. Only used, if there is no
        dedicated :attr:`.no_input_hint` for this element.

        Notes:
            The property searches for an option of the form
            ``<element_class>_debug`` in the section *debug* of
            config.conf. If no option is found, the return value is
            "You need to enter something".
        """
        name = f"no_input{type(self).__name__}"
        return self.experiment.config.get("hints", name, fallback="You need to enter something.")

    @property
    def prefix(self):
        """
        Union[str, Element]: A string or element, serving as prefix.

        If the prefix is an element, the getter returns only its
        inner html. If it is a string, the getter
        returns the text, wrapped in an appropriate html container.
        """
        try:
            return self._prefix._inner_html
        except AttributeError:
            return self._render_input_group_text(self._prefix)

    @property
    def suffix(self):
        """
        Union[str, Element]: A string or element, serving as suffix.

        If the suffix is an element, the getter returns only its
        inner html. If it is a string, the getter
        returns the text, wrapped in an appropriate html container.
        """
        try:
            return self._suffix._inner_html
        except AttributeError:
            return self._render_input_group_text(self._suffix)

    def _render_input_group_text(self, text: str) -> str:
        if text is not None:
            return f"<div class='input-group-text'>{text}</div>"
        else:
            return None

    @property
    def input(self) -> str:
        """
        str: Subject input to this element.
        """
        return self._input

    @input.setter
    def input(self, value):
        self._input = value

    @property
    def data(self) -> dict:
        """
        dict: Dictionary of element data.

        Includes the subject :attr:`.input` and the element's
        :attr:`.codebook_data`.
        """
        data = {}
        data["value"] = self.input
        data.update(self.codebook_data)
        return {self.name: data}

    def set_data(self, d: dict):
        """
        Sets the :attr:`.input` data.

        Args:
            d: A dictionary with data.

        Notes:
            The *d* dictionary will usually be a dictionary of all data
            collected on a page.
        """
        if not self.disabled:
            try:
                self.input = d[self.name]
            except KeyError:
                self.log.debug(f"No data for {self} found in data dictionary. Moving on.")
                pass

    @property
    def codebook_data(self) -> dict:
        """
        dict: Information about the element in dictionary form.
        """

        from alfred3 import page

        data = {}
        data["name"] = self.name
        data["label_top"] = self.toplab.text if self.toplab is not None else ""
        data["label_left"] = self.leftlab.text if self.leftlab is not None else ""
        data["label_right"] = self.rightlab.text if self.rightlab is not None else ""
        data["label_bottom"] = self.bottomlab.text if self.bottomlab is not None else ""
        data["tree"] = self.short_tree
        data["page_title"] = self.page.title
        data["element_type"] = type(self).__name__
        data["force_input"] = self._force_input
        data["prefix"] = self.prefix
        data["suffix"] = self.suffix
        data["default"] = self.default
        data["description"] = self.description
        data["unlinked"] = True if isinstance(self.page, page.UnlinkedDataPage) else False
        return data

    def added_to_page(self, page):

        from .. import page as pg

        if not isinstance(page, pg._PageCore):
            raise TypeError()

        self.page = page
        if self.name is None:
            raise ValueError(f"{self} is not named. Input elements must be named")

        if page.prefix_element_names:
            self.name = f"{self.page.name}_{self.name}"

        if self.page.experiment and not self.experiment:
            self.added_to_experiment(self.page.experiment)
        elif self.experiment:
            if self.name in self.experiment.root_section.all_updated_elements:
                raise AlfredError(
                    f"Element name '{self.name}' is already present in the experiment."
                )

        for fix in (self._prefix, self._suffix):
            try:
                fix.should_be_shown = False
                self.page += fix
            except AttributeError as e:
                self.log.debug(f"Exception passed silently: {e}")
                pass


@dataclass
class _Choice:
    """Dataclass for managing choices."""

    label: str = None
    label_id: str = None
    id: str = None
    name: str = None
    value: str = None
    type: str = "radio"
    checked: bool = False
    css_class: str = None
    disabled: bool = False


@inherit_kwargs(from_=[InputElement])
class ChoiceElement(InputElement, ABC):
    """
    Baseclass for derivation of choice elements.

    Args:
        *choice_labels: Variable number of strings or suitable elements
            to be used as choice labels. The number of labels determines
            the number of choices. We don't enforce harsh restrictions
            on the types of elements that can be used as choice labels,
            so it is up to you to make sensible choices. A common case
            would be an :class:`.Image`.
        vertical: Boolean switch, indicating whether the choices should
            be listed vertically. Defaults to *False*, i.e. horizontal
            display.

        {kwargs}

    """

    # Documented at :class:`.Element`
    element_template = jinja_env.get_template("html/ChoiceElement.html.j2")

    #: Choice type (e.g. "radio" for radio inputs and "checkbox")
    #: for multiple choice inputs
    type: str = None

    #: Switch for turning the interpretation of emoji shortcodes in the
    #: choice labels off, if necessary. Defaults to *True*.
    emojize: bool = True

    def __init__(
        self,
        *choice_labels: Union[str, Element],
        vertical: bool = False,
        align: str = "center",
        **kwargs,
    ):
        super().__init__(align=align, **kwargs)

        self._input = {}

        self.choice_labels = choice_labels  # documented in getter
        self.vertical = vertical  # documented in getter

        #: List of choices that belong to this element.
        self.choices: List[_Choice] = None

    @property
    def vertical(self):
        """
        bool: Attribute defining, whether the element is displayed vertically.
        """
        return self._vertical

    @vertical.setter
    def vertical(self, value):
        self._vertical = value

    @property
    def choice_labels(self):
        """
        list: Stored list of choice labels.
        """
        return self._choice_labels

    @choice_labels.setter
    def choice_labels(self, value: list):
        self._choice_labels = value

    def added_to_page(self, page):
        """
        If choice labels are element instances, they are added to the
        page to enable their full functionality.

        :meta private: (documented at :class:`.InputElement`)
        """
        super().added_to_page(page)

        for label in self.choice_labels:
            if isinstance(label, Element):
                label.added_to_page(page)
                label.should_be_shown = False
                label.width = "full"  # in case of TextElement, b/c its default is a special width

    def prepare_web_widget(self):

        self.choices = self.define_choices()

    @property
    def template_data(self):

        d = super().template_data
        d["choices"] = self.choices
        d["vertical"] = self.vertical
        d["type"] = self.type
        return d

    @abstractmethod
    def define_choices(self) -> List[_Choice]:
        """
        Abstract method for the definition of the individual choices
        belonging to this element.

        Must be redefined by inheriting elements and return a list
        of :class:`.Choice` instances.
        """
        pass

    @property
    def codebook_data(self):

        d = super().codebook_data

        for i, lab in enumerate(self.choice_labels, start=1):
            try:
                d.update({f"choice{i}": lab.text})  # if there is a text attribute, we use it.
            except AttributeError:
                d.update({f"choice{i}": str(lab)})  # otherwise __str__

        return d

    @property
    def input(self) -> dict:
        """
        Dict[str, bool]: Dictionary of subject inputs.

        Contains the choice labels as keys and their selection status
        (*True* for selected choices, *False* otherwise) as values.
        """
        return self._input

    @input.setter
    def input(self, value):
        self._input = value


# -*- coding: utf-8 -*-

"""
.. moduleauthor:: Paul Wiemann <paulwiemann@gmail.com>

Das Modul *ui_controller* stellt die Klassen zur Verfügung, die die Darstellung und die Steuerelemente auf verschiedenen Interfaces verwalten.
"""
import os
import threading
import time

from abc import ABCMeta, abstractmethod
from io import StringIO
from pathlib import Path
from uuid import uuid4
from dataclasses import dataclass
from typing import Union, Tuple

import importlib.resources
from jinja2 import Environment, PackageLoader

from .alfredlog import QueuedLoggingInterface
from .static import js
from .static import css
from .static import img
from . import element as elm
from .exceptions import AlfredError, ValidationError, AbortMove
from .util import icon

jinja_env = Environment(loader=PackageLoader("alfred3", "templates"))

@dataclass
class Move:
    """
    Template for saving data about participant movements.
    
    Attributes:
        exp_title: Experiment title
        exp_author: Experiment author
        exp_version: Experiment version
        exp_id: Experiment id
        exp_session_id: Experiment session id

        move_number: A continuously increasing counter of moves. Starts
            at one.
        tree: The full hierarchy of sections, indicating the exact 
            position of the Move's *page* in the experiment.
        page_name: The name of the Move's page.
        
        show_time: Time when the page was shown in seconds since epoch.
        hide_time: Time when the page was hidden in seconds since epoch.
        duration: Duration in seconds, calculated as the difference 
            hidetime - showtime.
        
        page_status_before_visit: Indicates the page's status before
            the visit represented by the Move instance, e.g. 'open', or 
            'closed'.
        page_status_before_visit: Indicates the page's status after
            the visit represented by the Move instance, e.g. 'open', or 
            'closed'.
        
        leave_in_direction: Indicates the direction, in which the page 
            was left (e.g. 'forward', 'backward', or 'jump').
        previous_page: Name of the previous page
        target_page: Name of the next page
        
        section_allows_backward: Indicates, whether participants can move
            backward from the Move's page.
        section_allows_forward: Indicates, whether participants can move
            forward from the Move's page.
        section_allows_jumpfrom: Indicates, whether participants can 
            jump from the Move's page.
        section_allows_jumpto: Indicates, whether participants can 
            jump to the Move's page.
    
    """
    exp_title: str = None
    exp_author: str = None
    exp_version: str = None
    exp_id: str = None
    exp_session_id: str = None
    
    move_number: int = None
    tree: str = None
    page_name: str = None
    
    show_time: float = None
    hide_time: float = None
    duration: float = None
    
    page_status_before_visit: str = None
    page_status_after_visit: str = None
    
    leave_in_direction: str = None
    previous_page: str = None
    target_page: str = None
    
    section_allows_forward: bool = None
    section_allows_backward: bool = None
    section_allows_jumpfrom: bool = None
    section_allows_jumpto: bool = None


class MovementManager:
    instance_log = False

    def __init__(self, experiment):
        self.exp = experiment
        self.experiment = experiment
        self.log = QueuedLoggingInterface(base_logger=__name__)
        self.log.add_queue_logger(self, __name__)

        self.current_index: int = 0
        self.previous_index: int = 0
        self.history: list = []
    
    @property
    def current_page(self):
        i = self.current_index
        return self.exp.root_section.all_pages_list[i]
    
    @property
    def previous_page(self):
        if self.current_page is self.first_page:
            return None
        
        i = self.previous_index
        return self.exp.root_section.all_pages_list[i]
    
    @property
    def next_page(self):
        if self.current_page is self.final_page:
            return None
        i = self.current_index + 1
        return self.exp.root_section.all_pages_list[i]
    
    @property
    def final_page(self):
        return self.exp.final_page
    
    def page_after(self, page):
        i = self.exp.root_section.all_page_names.index(page.name) + 1
        return self.exp.root_section.all_pages_list[i]
    
    def page_before(self, page):
        if self.current_page is self.first_page:
            return None
        i = self.exp.root_section.all_page_names.index(page.name) - 1
        return self.exp.root_section.all_pages_list[i]

    def find_page(self, query: Union[str, int]):
        """
        Find a page.

        Args:
            query: Can be either a page name or a page index.
        """
        page = self.experiment.root_section.all_pages.get(query, None)
        if page is not None:
            return page
        else:
            try:
                page = self.exp.root_section.all_pages_list[int(query)]
                return page
            except (IndexError, TypeError, ValueError):
                return None

    def index_of(self, page):
        return self.experiment.root_section.all_page_names.index(page.name)

    @property
    def first_page(self):
        return self.exp.root_section.all_pages_list[0]
    
    @property
    def first_visible_page(self):
        i = 0
        pages = self.exp.root_section.all_pages_list
        while not pages[i].should_be_shown:
            i += 1
        return pages[i]
    
    @property
    def last_page(self):
        return self.exp.root_section.all_pages_list[-2]
    
    def _abort_move(self):
        return self.current_index, self.current_index
    
    def _skip_page(self, to_page, direction: str) -> Tuple[int, int]:
        """
        Returns:
            Tuple[int, int]: A tuple of the current page and the target 
                page.

        Raises:
            AbortMove: If the movement is a jump and the target page
                (*to_page*) should not be shown.
        """
        if direction == "forward":
            self.log.debug(f"{to_page} should not be shown. Skipping page in direction 'forward'.")
            return self._move(to_page=self.page_after(to_page), direction="forward")
    
        elif direction == "backward":
            self.log.debug(f"{to_page} should not be shown. Skipping page in direction 'backward'.")
            return self._move(to_page=self.page_before(to_page), direction="backward")
    
        elif direction == "jump":
            self.log.debug(f"{to_page} should not be shown. Aborting move.")
            if self.exp.config.getboolean("general", "debug"):
                self.exp.message_manager.post_message(f"{to_page} should not be shown. Jump was aborted.", level="debug")
            raise AbortMove()
    
    def _check_permissions(self, to_page, direction: str):
        # check section permissions for jumps
        if direction == "jump":
            jump_allowed = self._check_jump_permission(next_page=to_page)
            if not jump_allowed:
                raise AbortMove()
        
        # check section permissions for normal moves
        elif not getattr(self.current_page.section, "allow_" + direction):
            self.log.debug(f"Section of page {self.current_page} does not allow movement in direction '{direction}'")
            raise AbortMove()
        
        # check section permission for target section
        elif direction == "backward" and not to_page.section.allow_backward:
            self.log.debug(f"Section of page {to_page} does not allow movement in direction '{direction}'")
            raise AbortMove()
    
    def _switch_sections(self, from_page, to_page, direction: str):
        page_status_before = from_page.is_closed
        
        if direction == "jump":
            from_page.section._move("jumpfrom")
        else:
            from_page.section._move(direction)

        if to_page.section.name in from_page.section.all_members:
            from_page.section._hand_over()
        else:
            from_page.section._leave()

        if self.exp.config.getboolean("data", "record_move_history"):
            self.record_move(page_status_before, direction=direction, to_page=to_page)
        
        if direction == "jump":
            to_page.section._move("jumpto")

        if from_page.section.name in to_page.section.all_members:
            to_page.section._resume()
        else:
            to_page.section._enter()
            
    def _conduct_movement(self, to_page, direction: str) -> Tuple[int, int]:
        """
        Raises:
            AbortMove: If section validation does not succeed.
        """
        current_page = self.current_page
        
        if not to_page.should_be_shown:
            # Executed here first for possibility of early cancellation
            return self._skip_page(to_page=to_page, direction=direction)
        
        # check section permissions for jumps
        self._check_permissions(to_page=to_page, direction=direction)

        self.log.debug(f"Moving from {self.current_page} to {to_page}, direction: '{direction}'.")

        # management of section entering and leaving behavior
        # and recording the move
        page_status_before = current_page.is_closed
        if current_page.section is not to_page.section:
            try:
                self._switch_sections(from_page=current_page, to_page=to_page, direction=direction)
            except ValidationError:
                raise AbortMove()
        
        else:
            try:
                if direction == "jump":
                    current_page.section._move("jumpfrom")
                    current_page.section._move("jumpto")
                else:
                    current_page.section._move(direction)
            except ValidationError:
                raise AbortMove()

            if self.exp.config.getboolean("data", "record_move_history"):
                self.record_move(page_status_before, direction=direction, to_page=to_page)

        now = time.time()
        to_page._on_showing_widget(show_time=now)

        if not to_page.should_be_shown:
        # executed here a second time, in case the to_page's 
        # should_be_shown attribute was changed during any of the move
        # functions

            return self._skip_page(to_page=to_page, direction=direction)
        
        return self.current_index, self.index_of(to_page)

    def _direct_jump(self, to):
        """
        Bypasses the normal movement system and permissions for a
        direct jump to the given page. Used, e.g. when an experiment is
        aborted.
        """
        to_page = self.find_page(query=to)
        self.log.debug(f"Direct jump to {to_page}. Skipping all normal move function.")
        if self.exp.config.getboolean("data", "record_move_history"):
                self.record_move(self.current_page.is_closed, direction="jump", to_page=to_page)
         
        self.current_index = self.index_of(to_page)

    def _move(self, to_page, direction: str) -> Tuple[int, int]:
        try:
            return self._conduct_movement(to_page=to_page, direction=direction)
        except AbortMove:
            return self._abort_move()
    
    def forward(self):
        to_page = self.next_page
        self.previous_index, self.current_index = self._move(to_page, "forward")
    
    def backward(self):
        to_page = self.page_before(self.current_page)
        self.previous_index, self.current_index = self._move(to_page, "backward")
    
    def _check_jump_permission(self, next_page):
        if self.experiment.config.getboolean("general", "debug"):
            self.log.debug("Debug mode enabled. Jump permission not checked.")
            self.experiment.message_manager.post_message("Debug mode enabled. Jump permission was not checked.", level="debug")
            return True

        elif not self.current_page is next_page and not self.current_page.section.allow_jumpfrom:
            self.log.debug(f"The section of page {self.current_page} cannot be jumped from. Aborting move.")
            msg = self.experiment.config.get("hints", "jumpfrom_forbidden")
            self.experiment.message_manager.post_message(msg, level="warning")
            return False

        elif not next_page.section.allow_jumpto:
            msg = self.experiment.config.get("hints", "jumpto_forbidden")
            self.log.debug(f"The section of page {next_page} cannot be jumped to. Aborting move.")
            self.experiment.message_manager.post_message(msg, level="warning")
            return False

        else:
            return True
    
    def jump(self, to: Union[str, int]):
        to_page = self.find_page(query=to)
        
        if not to_page:
            msg = self.experiment.config.get("hints", "jump_page_not_found")
            self.experiment.message_manager.post_message(msg, level="warning")
            return False
        
        self.previous_index, self.current_index = self._move(to_page=to_page, direction="jump")

    def record_move(self, page_was_closed: bool, direction: str, to_page):
        current_page = self.current_page
        
        move = Move()
        move.exp_title = self.exp.title
        move.exp_author = self.exp.author
        move.exp_version = self.exp.config.get("metadata", "version")
        move.exp_id = self.exp.exp_id
        move.exp_session_id = self.exp.session_id
        move.move_number = len(self.history) + 1
        move.tree = current_page.short_tree
        move.page_name = current_page.name
        move.page_status_before_visit = "closed" if page_was_closed else "open"
        move.page_status_after_visit = "closed" if current_page.is_closed else "open"
        move.show_time = current_page.show_times[-1]
        
        try:
            move.hide_time = current_page.hide_times[-1]
            move.duration = move.hide_time - move.show_time
        except IndexError:
            self.log.debug(f"Hide time for {current_page} not available. This may be normal for an abort page.")
            move.hide_time = None
            move.duration = None
        
        move.previous_page = self.previous_page.name if self.previous_page else None
        move.section_allows_forward = current_page.section.allow_forward
        move.section_allows_backward = current_page.section.allow_backward
        move.section_allows_jumpfrom = current_page.section.allow_jumpfrom
        move.section_allows_jumpto = current_page.section.allow_jumpto
        move.leave_in_direction = direction
        move.target_page = to_page.name
        self.history.append(move)

    def move(self, direction, to: Union[str, int] = None):

        if self.exp.session_expired:
            return self.exp.abort(
                reason="session timed out", 
                title="Session expired",
                icon="hourglass-end",
                msg= "Sorry! The experiment session timed out."
                )
        
        proceed = self.current_page.custom_move()
        now = time.time()
        try:
            self.current_page._on_hiding_widget(hide_time=now)
        except AbortMove:
            return
        
        if not proceed:
            self.log.debug(f"Page defined its own move method. Alfred's move system stands by.")
            return
        
        if self.exp.aborted:
            self.log.debug(f"A move was called, but the experiment is aborted. No move is being conducted.")
            return
        
        if direction == "stay":
            self.log.debug("Movement direction was 'stay' - no move conducted.")
            return
        elif direction == "forward":
            self.forward()
        elif direction == "backward":
            self.backward()
        elif direction == "jump":
            self.jump(to=to)
        elif direction.startswith("jump"):
            self.jump(to=direction[5:]) # jump string has the form 'jump>pagename'
    
    def start(self):

        self.current_index = self.index_of(self.first_visible_page)
        self.exp.root_section._enter()
        self.current_page._on_showing_widget(show_time=time.time())


class UserInterface:
    instance_log = False

    _css_files = [
        "bootstrap-4.5.3.min.css",
        "prism.css",
        "responsive.css",
    ]

    _js_files = [
        "jquery-3.5.1.min.js",
        "popper.min.js",
        "bootstrap-4.5.3.min.js",
        "detect.min.js",
        "prism.js",
        "font-awesome-icons.js",
        "responsive.js",

    ]

    def __init__(self, experiment):
        self.template = jinja_env.get_template("page.html.j2")

        self.experiment = experiment
        self.log = QueuedLoggingInterface(base_logger=__name__)
        self.log.add_queue_logger(self, __name__)

        self._basepath = self.experiment.config.get("webserver", "basepath")
        self._static_files = {}
        self._dynamic_files = {}
        self._callables = {}

        self.config = {}
        self.config["responsive"] = self.experiment.config.getboolean("layout", "responsive")
        self.config["website_title"] = self.experiment.config.get("layout", "website_title")
        self.config["logo_text"] = self.experiment.config.get("layout", "logo_text")
        self.config["footer_text"] = self.experiment.config.get("layout", "footer_text")

        with importlib.resources.path(img, "alfred_logo_color.png") as p:
            self.config["alfred_logo_url"] = self.add_static_file(p, content_type="image/png")

        self.css_urls = []
        self.js_urls = []

        self.css_code = []
        self.js_code = []

        
        self._determine_style()


        self._callables["clientinfo"] = self.save_client_info
        self._callables["set_page_data"] = self._set_page_data

        self.forward_enabled = True
        self.backward_enabled = True
        self.finish_enabled = True
        
        # the code block below enables the creation of standalone alfred3 html pages,
        # which don't host their own JavaScript and CSS on a localserver,
        # but instead place it directly in the html file.
        debug = self.experiment.config.getboolean("general", "debug")
        code_in_template = self.experiment.config.getboolean("debug", "code_in_templates")
        if debug and code_in_template:
            self._add_resources(self._js_files, "js")
            self._add_resources(self._css_files, "css")
        else:
            self._add_resource_links(self._js_files, "js")
            self._add_resource_links(self._css_files, "css")

    def _set_page_data(self, **data):
        data.pop("page_token", None)
        self.exp.current_page._set_data(data)
        self.log.info("Page data set via 'set_page_data' callable route.")
        self.exp.current_page.save_data()

    @property
    def exp(self):
        return self.experiment

    def _determine_style(self):
        """Adds .css styles and logo image to the layout.
        
        This method manages the switch between the two builtin styles 
        'base' and 'goe', and a possible customly defined style. If 
        reads the option "style" in the section "layout" of config.conf.
        """
        style = self.experiment.config.get("layout", "style")

        if style == "base":
            with importlib.resources.path(css, "base.css") as f:
                url = self.add_static_file(f, content_type="text/css")
                self.css_urls.append((5, url))
            
            self.config["logo_text"] = self.experiment.config.get("layout_base", "logo_text")
            logo = self.exp.subpath(self.exp.config.get("layout_goe", "logo"))
            if logo.is_file():
                self.config["logo_url"] = self.add_static_file(logo, content_type="image/png")

        elif style == "goe":

            with importlib.resources.path(css, "goe.css") as f:
                url = self.add_static_file(f, content_type="text/css")
                self.css_urls.append((5, url))

            with importlib.resources.path(img, "uni_goe_logo_white.png") as p:
                url = self.add_static_file(p, content_type="image/png")
                self.config["logo_url"] = url
            
            self.config["logo_text"] = self.experiment.config.get("layout_goe", "logo_text")

            logo = self.exp.subpath(self.exp.config.get("layout_goe", "logo"))
            if logo.is_file():
                self.config["logo_url"] = self.add_static_file(logo, content_type="image/png")
        else:
            raise ValueError(f"Invalid value for option 'style' in config section 'layout': {style}")
        
        # read CSS files from static folder
        static_folder = self.exp.config.get("layout", "static_folder")
        static_folder = self.exp.subpath(static_folder)
        try:
            for filename in static_folder.iterdir():
                if filename.is_file() and filename.suffix == ".css":
                    path = filename.resolve()
                    url = self.add_static_file(path, content_type="text/css")
                    self.css_urls.append((7, url))
        except FileNotFoundError:
            self.log.debug(f"Did not find static folder {static_folder}. Passing silently.")

        # read custom logo
        logo = self.experiment.config.get("layout", "logo")
        logo = self.experiment.subpath(logo)
        if logo.is_file():
            if logo.suffix == ".png":
                content_type = "image/png"
            elif logo.suffix in [".jpg", ".jpeg"]:
                content_type = "image/jpeg"
            logo_url = self.add_static_file(logo, content_type=content_type)
            self.config["logo_url"] = logo_url

    def _add_resource_links(self, resources: list, resource_type: str):
        """Adds resources to the UI via add_static_file.
        
        Args:
            resources: A list of tuples of the form (pkg, resource).
            resource_typetype: A string indicating the type of resource. 
                "js" for JavaScript, "css" for Cascading Style Sheets.
        """
        if resource_type == "js":
            container = self.js_urls
            pkg = js
        elif resource_type == "css":
            container = self.css_urls
            pkg = css

        for i, f in enumerate(resources):
            with importlib.resources.path(pkg, f) as p:
                url = self.add_static_file(p)
                container.append((i, url))

    def save_client_info(self, **data):
        """Updates the client info dictionary and saves data."""

        self.experiment.data_manager.client_data.update(data)
        self.experiment.movement_manager.current_page.save_data()

    def _add_resources(self, resources: list, resource_type: str):

        if resource_type == "js":
            container = self.js_code
            pkg = js
        elif resource_type == "css":
            container = self.css_code
            pkg = css

        for i, f in enumerate(resources):
            container.append((i, importlib.resources.read_text(pkg, f)))

    def code(self, page) -> dict:
        """Wraps the basic layout CSS and JavaScript together with
        the page's CSS and JavaScript in a single dictionary
        for easy use.
        """

        code = {}

        code["layout_css"] = sorted(self.css_urls)
        code["layout_js"] = sorted(self.js_urls)
        code["layout_css_code"] = sorted(self.css_code)
        code["layout_js_code"] = sorted(self.js_code)

        code["css_urls"] = page._css_urls
        code["css_code"] = page._css_code
        code["js_urls"] = page._js_urls
        code["js_code"] = page._js_code

        # JS Code for a single data saving call upon a visit to the first page
        # This is necessary in order to also save the screen resolution
        first_page = self.exp.movement_manager.first_visible_page
        if page is first_page and self.experiment.config.getboolean("data", "save_client_info"):
            code["js_code"] += [(7, importlib.resources.read_text(js, "clientinfo.js"))]
        
        try:
            code["css_code"] += self.exp.progress_bar.css_code
            code["js_code"] += self.exp.progress_bar.js_code
        except AttributeError:
            pass

        return code

    def render(self, page_token):
        """Renders the current page."""

        page = self.experiment.movement_manager.current_page
        d = {**self.config}
        
        if self.exp.config.getboolean("general", "debug") and not page is self.exp.final_page:
            d["debug"] = self.exp.config.getboolean("general", "debug")
            d["jumplist"] = page.elements[page.name + "__debug_jumplist__"]
        
        page.prepare_web_widget()

        d["code"] = self.code(page=page)
        d["page_token"] = page_token
        d["elements"] = page.elements.values()

        d["title"] = page.title
        d["subtitle"] = page.subtitle
        
        if not page is self.exp.movement_manager.first_page:
            previous_section = self.exp.movement_manager.page_before(page).section
            if previous_section.allow_backward and page.section.allow_backward:
                d["backward_text"] = self.experiment.config.get("navigation", "backward")
        
        if page.section.allow_forward:
            if self.exp.movement_manager.next_page is self.exp.root_section.final_page:
                d["finish_text"] = self.experiment.config.get("navigation", "finish")
            else:
                d["forward_text"] = self.experiment.config.get("navigation", "forward")

        messages = self.experiment.message_manager.get_messages()
        if messages:
            d["messages"] = messages

        # progress bar
        d["show_progress"] = self.experiment.config.getboolean("layout", "show_progress")
        d["fix_progress_top"] = self.experiment.config.getboolean("layout", "fix_progress_top")
        if d["show_progress"]:
            self.exp.progress_bar._prepare_web_widget()
            d["progress"] = self.exp.progress_bar.web_widget

        return self.template.render(d)

    def render_html(self, page_token):
        """Alias for render, provided for compatibility."""
        return self.render(page_token=page_token)

    @property
    def basepath(self):

        if self._basepath is not None:
            return self._basepath
        else:
            return ""

    def get_static_file(self, identifier):
        """Returns the filepath to a static file based on its unique ID.

        Args:
            identifier: Unique ID of a static file.
        """
        return self._static_files[identifier]

    def add_static_file(self, path, content_type=None):
        """Adds a static file to an internal list. This allows us to
        keep the actual filepath private, which is a security feature
        for web experiments.

        Returns the anonymized url for the added file.

        Args:
            path: Path to file.
            content_type: Mimetype of the added file.
        """

        path = Path(path)
        if not path.is_absolute():
            path = self.experiment.path / path
        
        if self.experiment.config.getboolean("general", "debug"):
            identifier = str(path.name)
        else:
            identifier = uuid4().hex

        # the code below causes alfred to fail to detect changes in
        # static files when debug mode is activated

        # if self.experiment and self.experiment.config.getboolean("general", "debug"):
        #     if not hasattr(self, "sf_counter"):
        #         self.sf_counter = 0
        #     self.sf_counter += 1
        #     identifier = str(self.sf_counter)

        self._static_files[identifier] = (path, content_type)

        url = f"{self.basepath}/staticfile/{identifier}"
        return url

    def get_dynamic_file(self, identifier):
        return self._dynamic_files[identifier]

    def add_dynamic_file(self, file_obj, content_type=None):
        identifier = uuid4().hex
        while identifier in self._dynamic_files:
            identifier = uuid4().hex

        self._dynamic_files[identifier] = (file_obj, content_type)
        url = "{basepath}/dynamicfile/{identifier}".format(
            basepath=self._basepath, identifier=identifier
        )
        return url

    def get_callable(self, identifier):
        return self._callables[identifier]

    def add_callable(self, f: callable):
        identifier = uuid4().hex
        while identifier in self._callables:
            identifier = uuid4().hex

        self._callables[identifier] = f
        url = "{basepath}/callable/{identifier}".format(
            basepath=self._basepath, identifier=identifier
        )
        return url

    def start(self):
        self.exp.movement_manager.start()

